/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 1999,2000 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 * if any, must include the following acknowledgment:
 * "This product includes software developed by the
 * Apache Software Foundation (http://www.apache.org/)."
 * Alternately, this acknowledgment may appear in the software itself,
 * if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Xerces" and "Apache Software Foundation" must
 * not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 * nor may "Apache" appear in their name, without prior written
 * permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 1999, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
package org.enhydra.xml.xhtml.dom.xerces;

/*
 * Derived from Xerces HTMLDocument implementation.
 * Author: Assaf Arkin <arkin@exoffice.com>
 */
import java.io.*;
import java.util.HashMap;
import java.lang.reflect.*;
import org.w3c.dom.*;
import org.w3c.dom.html.*;
import org.enhydra.apache.xerces.dom.DocumentImpl;
import org.enhydra.apache.xerces.dom.NodeImpl;
import org.enhydra.apache.xerces.dom.AttrImpl;
import org.enhydra.xml.xmlc.XMLObjectLink;
import org.enhydra.xml.xmlc.XMLObject;

/**
 * Implements an XHTML document. Provides access to the top level element in
 * the document, its body and title.
 *
 * @see org.w3c.dom.html.HTMLDocument
 */
public class XHTMLDocumentImpl extends XHTMLDocumentBase implements HTMLDocument, XMLObjectLink {
    /**
     * Holds names and classes of HTML element types. When an element with a
     * particular tag name is created, the matching {@link java.lang.Class}
     * is used to create the element object. For example, &lt;A&gt; matches
     * {@link HTMLAnchorElementImpl}. This static table is shared across all
     * HTML documents.
     *
     * @see #createElement
     */
    private static HashMap fElementTypes;

    /**
     * Constructor without document type.
     */
    public XHTMLDocumentImpl() {
	super(fElementTypes);
    }

    /**
     * Constructor with document type.
     */
    public XHTMLDocumentImpl(DocumentType doctype) {
	super(doctype, fElementTypes);
    }
    /**
     * Class initializer.
     */
    static {
	fElementTypes = new HashMap();
        fElementTypes.put("basefont", XHTMLBaseFontElementImpl.class);
        fElementTypes.put("button", XHTMLButtonElementImpl.class);
        fElementTypes.put("textarea", XHTMLTextAreaElementImpl.class);
        fElementTypes.put("em", XHTMLEmElementImpl.class);
        fElementTypes.put("small", XHTMLSmallElementImpl.class);
        fElementTypes.put("area", XHTMLAreaElementImpl.class);
        fElementTypes.put("noframes", XHTMLNoframesElementImpl.class);
        fElementTypes.put("bdo", XHTMLBdoElementImpl.class);
        fElementTypes.put("form", XHTMLFormElementImpl.class);
        fElementTypes.put("link", XHTMLLinkElementImpl.class);
        fElementTypes.put("label", XHTMLLabelElementImpl.class);
        fElementTypes.put("dt", XHTMLDtElementImpl.class);
        fElementTypes.put("span", XHTMLSpanElementImpl.class);
        fElementTypes.put("isindex", XHTMLIsIndexElementImpl.class);
        fElementTypes.put("title", XHTMLTitleElementImpl.class);
        fElementTypes.put("strong", XHTMLStrongElementImpl.class);
        fElementTypes.put("script", XHTMLScriptElementImpl.class);
        fElementTypes.put("div", XHTMLDivElementImpl.class);
        fElementTypes.put("dl", XHTMLDListElementImpl.class);
        fElementTypes.put("blockquote", XHTMLQuoteElementImpl.class);
        fElementTypes.put("kbd", XHTMLKbdElementImpl.class);
        fElementTypes.put("menu", XHTMLMenuElementImpl.class);
        fElementTypes.put("body", XHTMLBodyElementImpl.class);
        fElementTypes.put("dir", XHTMLDirectoryElementImpl.class);
        fElementTypes.put("ins", XHTMLModElementImpl.class);
        fElementTypes.put("map", XHTMLMapElementImpl.class);
        fElementTypes.put("dd", XHTMLDdElementImpl.class);
        fElementTypes.put("fieldset", XHTMLFieldSetElementImpl.class);
        fElementTypes.put("head", XHTMLHeadElementImpl.class);
        fElementTypes.put("col", XHTMLTableColElementImpl.class);
        fElementTypes.put("base", XHTMLBaseElementImpl.class);
        fElementTypes.put("big", XHTMLBigElementImpl.class);
        fElementTypes.put("meta", XHTMLMetaElementImpl.class);
        fElementTypes.put("code", XHTMLCodeElementImpl.class);
        fElementTypes.put("tbody", XHTMLTbodyElementImpl.class);
        fElementTypes.put("option", XHTMLOptionElementImpl.class);
        fElementTypes.put("u", XHTMLUElementImpl.class);
        fElementTypes.put("s", XHTMLSElementImpl.class);
        fElementTypes.put("q", XHTMLQuoteElementImpl.class);
        fElementTypes.put("p", XHTMLParagraphElementImpl.class);
        fElementTypes.put("ol", XHTMLOListElementImpl.class);
        fElementTypes.put("thead", XHTMLTheadElementImpl.class);
        fElementTypes.put("ul", XHTMLUListElementImpl.class);
        fElementTypes.put("i", XHTMLIElementImpl.class);
        fElementTypes.put("pre", XHTMLPreElementImpl.class);
        fElementTypes.put("optgroup", XHTMLOptGroupElementImpl.class);
        fElementTypes.put("img", XHTMLImageElementImpl.class);
        fElementTypes.put("caption", XHTMLTableCaptionElementImpl.class);
        fElementTypes.put("b", XHTMLBElementImpl.class);
        fElementTypes.put("a", XHTMLAnchorElementImpl.class);
        fElementTypes.put("frame", XHTMLFrameElementImpl.class);
        fElementTypes.put("br", XHTMLBRElementImpl.class);
        fElementTypes.put("style", XHTMLStyleElementImpl.class);
        fElementTypes.put("hr", XHTMLHRElementImpl.class);
        fElementTypes.put("param", XHTMLParamElementImpl.class);
        fElementTypes.put("table", XHTMLTableElementImpl.class);
        fElementTypes.put("applet", XHTMLAppletElementImpl.class);
        fElementTypes.put("tt", XHTMLTtElementImpl.class);
        fElementTypes.put("tr", XHTMLTableRowElementImpl.class);
        fElementTypes.put("th", XHTMLTableCellElementImpl.class);
        fElementTypes.put("center", XHTMLCenterElementImpl.class);
        fElementTypes.put("td", XHTMLTableCellElementImpl.class);
        fElementTypes.put("samp", XHTMLSampElementImpl.class);
        fElementTypes.put("tfoot", XHTMLTfootElementImpl.class);
        fElementTypes.put("font", XHTMLFontElementImpl.class);
        fElementTypes.put("dfn", XHTMLDfnElementImpl.class);
        fElementTypes.put("noscript", XHTMLNoscriptElementImpl.class);
        fElementTypes.put("object", XHTMLObjectElementImpl.class);
        fElementTypes.put("colgroup", XHTMLTableColElementImpl.class);
        fElementTypes.put("sup", XHTMLSupElementImpl.class);
        fElementTypes.put("html", XHTMLHtmlElementImpl.class);
        fElementTypes.put("h6", XHTMLHeadingElementImpl.class);
        fElementTypes.put("h5", XHTMLHeadingElementImpl.class);
        fElementTypes.put("h4", XHTMLHeadingElementImpl.class);
        fElementTypes.put("h3", XHTMLHeadingElementImpl.class);
        fElementTypes.put("frameset", XHTMLFrameSetElementImpl.class);
        fElementTypes.put("h2", XHTMLHeadingElementImpl.class);
        fElementTypes.put("h1", XHTMLHeadingElementImpl.class);
        fElementTypes.put("iframe", XHTMLIFrameElementImpl.class);
        fElementTypes.put("strike", XHTMLStrikeElementImpl.class);
        fElementTypes.put("sub", XHTMLSubElementImpl.class);
        fElementTypes.put("acronym", XHTMLAcronymElementImpl.class);
        fElementTypes.put("select", XHTMLSelectElementImpl.class);
        fElementTypes.put("del", XHTMLModElementImpl.class);
        fElementTypes.put("li", XHTMLLIElementImpl.class);
        fElementTypes.put("cite", XHTMLCiteElementImpl.class);
        fElementTypes.put("var", XHTMLVarElementImpl.class);
        fElementTypes.put("legend", XHTMLLegendElementImpl.class);
        fElementTypes.put("abbr", XHTMLAbbrElementImpl.class);
        fElementTypes.put("input", XHTMLInputElementImpl.class);
        fElementTypes.put("address", XHTMLAddressElementImpl.class);;
    }

    //-------------------------------------------------------------------------
    // XMLObjectLink implementation
    //-------------------------------------------------------------------------

    /**
     * Reference to XMLObject containing this Document.
     */
    private XMLObject fXmlObjectLink;

    /**
     * @see XMLObjectLink#setXMLObject
     */
    public void setXMLObject(XMLObject xmlObject) {
        fXmlObjectLink = xmlObject;
    }

    /**
     * @see XMLObjectLink#getXMLObject
     */
    public XMLObject getXMLObject() {
        return fXmlObjectLink;
    }

}
