/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: LazyNotation.java,v 1.2 2003/06/29 05:58:48 jkjome Exp $
 */

package org.enhydra.xml.lazydom;

import org.w3c.dom.Document;
import org.w3c.dom.Notation;
import org.w3c.dom.Node;
import org.enhydra.apache.xerces.dom.NotationImpl;

/**
 * Lazy notation node.
 */
public class LazyNotation extends NotationImpl implements LazyNode {
    /**
     * Constructor
     * @param ownerDoc The document that owns this node.
     * @param template If not-null, get the parameters from this template.
     * @param name Document type name.
     *  Will be ignored if template is not-null.
     * @param publicId The public id.
     *  Will be ignored if template is not-null.
     * @param systemId The system id.
     *  Will be ignored if template is not-null.
     */
    protected LazyNotation(LazyDocument ownerDoc,
                           LazyNotation template,
                           String name,
                           String publicId,
                           String systemId) {
    	super(ownerDoc, 
              (template != null) ? template.getNodeName() : name);
        if (template != null) {
            if (template.getPublicId() != null) {
                setPublicId(template.getPublicId());
            }
            if (template.getSystemId() != null) {
                setSystemId(template.getSystemId());
            }
            fTemplateNode = template;
            fNodeId = template.getNodeId();
        } else {
            if (publicId != null) {
                setPublicId(publicId);
            }
            if (systemId != null) {
                setSystemId(systemId);
            }
        }
    }

    //-------------------------------------------------------------------------
    // LazyNotation specific
    //-------------------------------------------------------------------------

    /**
     * Template for this Notation.
     */
    private LazyNotation fTemplateNode = null;

    /**
     * Get the template for this node.
     * @see LazyNode#getTemplateNode
     */
    public LazyNotation getTemplateNotation() {
        return fTemplateNode;
    }

    /**
     * @see Node#cloneNode
     */
    public Node cloneNode(boolean deep) {
        // This does a clone(), must clean up all fields.
        LazyNotation newNotation = (LazyNotation)super.cloneNode(deep);
        newNotation.fNodeId = NULL_NODE_ID;
        return newNotation;
    }

    //-------------------------------------------------------------------------
    // LazyNode support
    //-------------------------------------------------------------------------

    /*
     * Node id for this element.
     */
    private int fNodeId = NULL_NODE_ID;

    /**
     * Is this a template node?
     */
    private boolean fIsTemplateNode;

    /*
     * @see LazyNode#makeTemplateNode
     */
    public void makeTemplateNode(int nodeId) {
        fNodeId = nodeId;
        fIsTemplateNode = true;
    }

    /**
     * @see LazyNode#getNodeId
     */
    public int getNodeId() {
        return fNodeId;
    }

    /**
     * @see LazyNode#isTemplateNode
     */
    public boolean isTemplateNode() {
        return fIsTemplateNode;
    }

    /**
     * @see LazyNode#getTemplateNode
     */
    public LazyNode getTemplateNode() {
        return fTemplateNode;
    }

    /**
     * @see LazyNode#templateClone
     */
    public LazyNode templateClone(Document ownerDocument) {
        return new LazyNotation((LazyDocument)ownerDocument, this,
                                null, null, null);
    }

    /**
     * Set the node value, invalidating the id.  All node data is modified
     * by this routine.
     * @see org.w3c.dom.Node#setNodeValue
     */
    public void setNodeValue(String value) {
        fNodeId = NULL_NODE_ID;
        super.setNodeValue(value);
    }

}
