/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: DocClassCmdOptions.java,v 1.3 2003/09/15 16:57:03 jkjome Exp $
 */

package org.enhydra.xml.xmlc.commands.xmlc;

import java.io.File;
import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.xmlc.commands.options.*;
import org.enhydra.xml.xmlc.metadata.*;
import org.enhydra.xml.io.ErrorReporter;

/**
 * Command line options parsing document class options.
 */
class DocClassCmdOptions extends BaseCmdOptions {
    /**
     * Class for the -implements option.
     */
    private class ImplementsOption extends Option {
        /**
         * Constructor.
         */
        public ImplementsOption() {
            super("-implements", 1, true,
                  "interface - Specify an interface the generate class will implement");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args,
                             ErrorReporter errorReporter,
                             Object clientData) throws XMLCException {
            ((MetaData)clientData).getDocumentClass().addImplements(args[0]);
        }
    }

    /**
     * Class for the -generate option.
     */
    private class GenerateOption extends Option {
        /**
         * Constructor.
         */
        public GenerateOption() {
            super("-generate", 1, true,
                  "generate which - Specify type of class to generate (both, interface, implementation, class)");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args,
                             ErrorReporter errorReporter,
                             Object clientData) throws XMLCException {
            GenerateType generate;
            try {
                generate = GenerateType.getType(args[0]);
            } catch (IllegalArgumentException except) {
                throw new XMLCException(except.getMessage(),
                                        except);
            }
            ((MetaData)clientData).getDocumentClass().setGenerate(generate);
        }
    }

    /**
     * Class for the -class option.
     */
    private class ClassOption extends Option {
        /**
         * Constructor.
         */
        public ClassOption() {
            super( "-class", 1, false,
                   "class - Fully qualified class name for the generated class or interface.");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args,
                             ErrorReporter errorReporter,
                             Object clientData) throws XMLCException {
            ((MetaData)clientData).getDocumentClass().setName(args[0]);
        }
    }

    /**
     * Class for the -extends option.
     */
    private class ExtendsOption extends Option {
        /**
         * Constructor.
         */
        public ExtendsOption() {
            super("-extends", 1, false,
                  "classname - Class the generated document extends");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args,
                             ErrorReporter errorReporter,
                             Object clientData) throws XMLCException {
            ((MetaData)clientData).getDocumentClass().setExtends(args[0]);
        }
    }

    /**
     * Class for the -for-recomp option.
     */
    private class ForRecompOption extends FlagOption {
        /**
         * Constructor.
         */
        public ForRecompOption() {
            super("-for-recomp",
                  "Deprecated.  No longer part of XMLC proper...too container (Enhydra3) specific.  Use -for-deferred-parsing instead.");
        }

        /**
         * Set value.
         */
        protected void set(Object clientData) throws XMLCException {
            ((MetaData)clientData).getDocumentClass().setRecompilation(true);
        }
    }

    /**
     * Class for the -for-deferred-parsing option.
     */
    private class ForDeferredParsingOption extends FlagOption {
        /**
         * Constructor.
         */
        public ForDeferredParsingOption() {
            super("-for-deferred-parsing",
                  "Generate support for deferred parsing and automatic re-parsing of the actual document *ML.");
        }

        /**
         * Set value.
         */
        protected void set(Object clientData) throws XMLCException {
            ((MetaData)clientData).getDocumentClass().setDeferredParsing(true);
        }
    }

    /**
     * Class for the -domfactory option.
     */
    private class DOMFactoryOption extends Option {
        /**
         * Constructor.
         */
        public DOMFactoryOption() {
            super("-domfactory", 1, false,
                  "classname - XMLCDOMFactory class for creating DTD-specific documents");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args,
                             ErrorReporter errorReporter,
                             Object clientData) throws XMLCException {
            DocumentClass docClass = ((MetaData)clientData).getDocumentClass();
            if (docClass.getDom() != null) {
                throw new XMLCException("can't specify both -dom and -domfactory");
            }
            docClass.setDomFactory(args[0]);
        }
    }

    /**
     * Class for the -dom option.
     */
    private class DOMOption extends Option {
        /**
         * Constructor.
         */
        public DOMOption() {
            super("-dom", 1, false,
                  "dom - Specify one of a predefined set of DOM factories; one of `lazydom', `xerces', 'generic', 'xhtml', 'wml', 'voicexml', or 'chtml'");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args,
                             ErrorReporter errorReporter,
                             Object clientData) throws XMLCException {
            DocumentClass docClass = ((MetaData)clientData).getDocumentClass();
            if (docClass.getDomFactory() != null) {
                throw new XMLCException("can't specify both -dom and -domfactory");
            }
            docClass.setDom(DOMType.getType(args[0]));
        }
    }

    /**
     * Class for the -create-get-tag-methods option.
     */
    private class CreateGetTagMethods extends FlagOption {
        /**
         * Constructor.
         */
        public CreateGetTagMethods() {
            super("-create-get-tag-methods",
                  "name - getTagXXX() methods with generic return types");
        }

        /**
         * Set value.
         */
        protected void set(Object clientData) throws XMLCException {
            ((MetaData)clientData).getDocumentClass().setCreateGetTagMethods(true);
        }
    }

    /**
     * Constructor.  Add options to option set.
     */
    public DocClassCmdOptions(OptionSet optionSet) {
        super(optionSet);
        optionSet.addOption(new ImplementsOption());
        optionSet.addOption(new GenerateOption());
        optionSet.addOption(new ClassOption());
        optionSet.addOption(new ExtendsOption());
        optionSet.addOption(new ForRecompOption());
        optionSet.addOption(new ForDeferredParsingOption());
        optionSet.addOption(new DOMFactoryOption());
        optionSet.addOption(new DOMOption());
        optionSet.addOption(new GenerateOption());
        optionSet.addOption(new CreateGetTagMethods());
    }
}
