/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: HTMLCmdOptions.java,v 1.1.1.1 2003/03/10 16:36:20 taweili Exp $
 */

package org.enhydra.xml.xmlc.commands.xmlc;

import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.xmlc.commands.options.*;
import org.enhydra.xml.xmlc.metadata.*;
import org.enhydra.xml.io.ErrorReporter;

/**
 * Command line options parsing for -html:* options.
 */
class HTMLCmdOptions extends BaseCmdOptions {
    /**
     * Class for the -html:encoding option.
     */
    private class HTMLEncodingOption extends Option {
        /**
         * Construct a new object.
         */
        public HTMLEncodingOption() {
            super("-html:encoding", 1, false,
                  "encoding - Specify the encoding to use when reading a HTML document.");
        }

        /**
         * Parse the option.
         */
        public void parse(String[] args,
                          ErrorReporter errorReporter,
                          Object clientData) throws XMLCException {
            ((MetaData)clientData).getHTMLSection().setEncoding(args[0]);
        }
    }

    /**
     * Parser for deprecated -html:frameset option.
     */
    private class HTMLFrameSetOption extends Option {
        /**
         * Constructor.
         */
        public HTMLFrameSetOption() {
            super("-html:frameset", 0, false,
                  "- Deprecated and ignored.");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args,
                             ErrorReporter errorReporter,
                             Object clientData) throws XMLCException {
            errorReporter.warning("The -html:frameset option is deprecated and ignored");
        }
    }

    /**
     * Class for -html:old-class-constants option
     */
    private class HTMLOldClassConstantsOption extends Option {
        /**
         * Constructor.
         */
        public HTMLOldClassConstantsOption() {
            super("-html:old-class-constants", 0, false,
                  "- Generate old-style, all upper-case class constants");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args,
                             ErrorReporter errorReporter,
                             Object clientData) throws XMLCException {
            HTMLCompatibility compat = ((MetaData)clientData).getHTMLSection().getCreateCompatibility();
            compat.setOldClassConstants(true);
            errorReporter.warning("The " + getName()
                                  + " option will be deprecated in a future release, please update your code");
        }
    }

    /**
     * Class for -html:old-name-constants option
     */
    private class HTMLOldNameConstantsOption extends Option {
        /**
         * Constructor.
         */
        public HTMLOldNameConstantsOption() {
            super("-html:old-name-constants", 0, false,
                  "- Generate old-style, all upper-case name constants");
        }

        /**
         * Parse an instance of the option.
         */
        protected void parse(String[] args,
                             ErrorReporter errorReporter,
                             Object clientData) throws XMLCException {
            HTMLCompatibility compat
                = ((MetaData)clientData).getHTMLSection().getCreateCompatibility();
            compat.setOldNameConstants(true);
            errorReporter.warning("The " + getName()
                                  + " option will be deprecated in a future release, please update your code");
        }
    }

    /**
     * Class for the -html:addtagset option.
     */
    private class HTMLAddTagSetOption extends Option {
        /**
         * Construct a new object.
         */
        public HTMLAddTagSetOption() {
            super("-html:addtagset", 1, true,
                  "tagsetname - Add a predefined set of proprietary tags to the list of valid HTML tags.");
        }

        /**
         * Parse an instance of the option and set the field in the Options
         * object using Java reflection.
         *
         * @param args The option's arguments.  Should have one.
         */
        public void parse(String[] args,
                          ErrorReporter errorReporter,
                          Object clientData) throws XMLCException {
            HTMLTagSet tagSet = ((MetaData)clientData).getHTMLSection().addHTMLTagSet();
            try {
                tagSet.setTagSet(HTMLTagSetType.getType(args[0]));
            } catch (IllegalArgumentException except) {
                throw new XMLCException(except.getMessage(),
                                        except);
            }
        }
    }

    /**
     * Class for the -html:addtag option.
     */
    private class HTMLAddTagOption extends Option {
        /**
         * Construct a new object.
         */
        public HTMLAddTagOption() {
            super("-html:addtag", 2, true,
                "tagname flags - Add a proprietary tag to the list of valid HTML tags.");
        }

        /**
         * Parse the flags representing the content model.
         */
        private void parseFlags(String flagsStr,
                                HTMLTag tag) throws XMLCException {
            // Split list at commas and convert to bit set.
            int startIdx = 0;
            while (startIdx < flagsStr.length()) {
                int nextIdx = flagsStr.indexOf(',', startIdx);
                if (nextIdx < 0) {
                    nextIdx = flagsStr.length(); // End of string
                }
                String flagStr = flagsStr.substring(startIdx, nextIdx);
                if (flagStr.equalsIgnoreCase("inline")) {
                    tag.setInline(true);
                } else if (flagStr.equalsIgnoreCase("block")) {
                    tag.setBlock(true);
                } else if (flagStr.equalsIgnoreCase("empty")) {
                    tag.setEmpty(true);
                } else if (flagStr.equalsIgnoreCase("opt")) {
                    tag.setOptclose(true);
                } else {
                    throw new XMLCException("Invaild flag for " + name
                                            + " \"" + flagStr + "\"");
                }
                startIdx = nextIdx+1;
            }
        }

        /**
         * Parse an instance of the option and set the field in the Options
         * object using Java reflection.
         *
         * @param args The option's arguments.  Should have one.
         */
        public void parse(String[] args,
                          ErrorReporter errorReporter,
                          Object clientData) throws XMLCException {
            HTMLTag tag = ((MetaData)clientData).getHTMLSection().addHTMLTag();
            tag.setName(args[0]);
            parseFlags(args[1], tag);
        }
    }

    /**
     * Class for the -html:addattr option.
     */
    private class HTMLAddAttrOption extends Option {
        /**
         * Construct a new object.
         */
        public HTMLAddAttrOption() {
            super("-html:addattr", 1, true,
                "attrname - Add a proprietary attribute the list of valid HTML attributes.");
        }

        /**
         * Parse an instance of the option and set the field in the Options
         * object using Java reflection.
         *
         * @param args The option's arguments.  Should have one.
         */
        public void parse(String[] args,
                          ErrorReporter errorReporter,
                          Object clientData) throws XMLCException {
            HTMLAttr attr = ((MetaData)clientData).getHTMLSection().addHTMLAttr();
            attr.setName(args[0]);
        }
    }

    /**
     * Constructor.  Add options to option set.
     */
    public HTMLCmdOptions(OptionSet optionSet) {
        super(optionSet);
        optionSet.addOption(new HTMLFrameSetOption());
        optionSet.addOption(new HTMLEncodingOption());
        optionSet.addOption(new HTMLAddTagSetOption());
        optionSet.addOption(new HTMLAddTagOption());
        optionSet.addOption(new HTMLAddAttrOption());
        optionSet.addOption(new HTMLOldClassConstantsOption());
        optionSet.addOption(new HTMLOldNameConstantsOption());
    }
}
