/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: LazyDOMDocBuilderGenerator.java,v 1.1.1.1 2003/03/10 16:36:22 taweili Exp $
 */

package org.enhydra.xml.xmlc.dom.lazydom;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.codegen.*;
import org.enhydra.xml.xmlc.dom.*;
import org.enhydra.xml.xmlc.dom.generic.NodeCreateGenerator;
import org.enhydra.xml.xmlc.dom.generic.DOMBuilderGenerator;
import org.enhydra.xml.xmlc.compiler.ClassGenerator;
import org.enhydra.xml.xmlc.compiler.ElementInfo;
import org.enhydra.xml.xmlc.compiler.ElementTable;
import org.enhydra.xml.lazydom.*;
import org.enhydra.xml.io.OutputOptions;
import org.enhydra.xml.io.PreFormattedTextDocument;
import java.util.HashMap;
import java.util.Iterator;
import java.util.ArrayList;
import org.w3c.dom.*;
import org.w3c.dom.html.HTMLDocument;

//FIXME: review use of document vs XMLCDocument.

/**
 * Class to generate code to build the document tree for the Lazy DOM.
 * This creates a static template DOM that is used to instantiate individual
 * nodes. 
 * <p>
 * The static template is initialized by a set of methods to get around the
 * method size limitation of doing this as an array initializer.
 */
public class LazyDOMDocBuilderGenerator implements DocBuilderGenerator {
    /*
     * Maximum creation-cost per build method.  One unit is approximately the
     * cost to create one node.
     */
    private static final int MAX_CREATE_COST_PER_BUILD_METHOD = 256;

    /**
     * Name of static template document table.
     */
    private static String TEMPLATE_DOCUMENT_FIELD = "fTemplateDocument";
    
    /*
     * Name of variable in init function that holds document type object.
     */
    private static final String DOC_TYPE_VAR = "documentType";

    /** 
     * Name for the variable holding the Document during creation.
     */
    public static final String DOCUMENT_VAR_NAME = "document";

    /**
     * Name of class field containing OutputOptions that was used
     * to preformat the object.
     */
    private static String PREFORMAT_OUTPUT_OPTIONS_FIELD = "fPreFormatOutputOptions";
    
    /**
     * Field containing LazyDocument.
     */
    static String LAZY_DOCUMENT_FIELD = "fLazyDocument";

    /**
     * Base name of template DOM building methods.
     */
    private static final String SUBDOCUMENT_BUILD_METHOD = "buildTemplateSubDocument";

    /**
     * Use a TemplateDOM object to assign node ids.  Ensures that
     * the node ids are identical.
     */
    private TemplateDOM templateDOM;

    /**
     * If not null, the object to preformat node text.
     */
    private PreFormatter fPreFormatter;

    /**
     * Generate code to create a document from a TemplateDOM.
     * @param document Document that will be recreated.
     * @param templateFieldName Name of template field.
     * @param docFieldName Name of field for document object.
     */
    public void generateDocCreateFromTemplate(Document document,
                                              String templateFieldName,
                                              String docFieldName,
                                              JavaCode body) {
        String docClass = document.getClass().getName();

        body.addln(docFieldName + " = (" + docClass + ")(("
                   + LazyDomFactory.class.getName()
                   + ")" + ClassGenerator.DOM_FACTORY_FIELD_NAME
                   + ").createDocument(" + templateFieldName + ");");
    }

    /**
     * Do initialization for the preformatter.
     */
    private void preFormatterInit(JavaClass docClass) {
        // Create static variable that will hold the OutputOptions used to
        // format the class
        docClass.addField(new JavaField(PREFORMAT_OUTPUT_OPTIONS_FIELD,
                                        OutputOptions.class.getName(),
                                        JavaModifiers.PRIVATE | JavaModifiers.STATIC | JavaModifiers.FINAL,
                                        new String[]{"Options used to preformat the document when compiled"},
                                        null));

        // Generate code to recreate the output options.
        fPreFormatter.createOutputOptionsCodeGenerator(PREFORMAT_OUTPUT_OPTIONS_FIELD,
                                                       docClass.getClassInitializer());
    }

    /**
     * Generate template document initialization,
     */
    private void createTemplateDocInit(XMLCDocument xmlcDoc,
                                       AccessorGenerator accessorGenerator,
                                       ElementTable elementTable,
                                       JavaClass docClass) throws XMLCException {
        LazyDocument document = (LazyDocument)xmlcDoc.getDocument();

        // This assigns node ids.
        templateDOM = new TemplateDOM(document);

        JavaCode body = docClass.getClassInitializer();
        DOMBuilderGenerator domBuilderGenerator
            = new DOMBuilderGenerator(SUBDOCUMENT_BUILD_METHOD,
                                      xmlcDoc.getDocument(),
                                      LazyDocument.class.getName(),
                                      new LazyDOMNodeCreateGenerator(xmlcDoc,
                                                                     fPreFormatter),
                                      accessorGenerator,
                                      elementTable,
                                      docClass,
                                      MAX_CREATE_COST_PER_BUILD_METHOD,
                                      true);

        JavaField field = new JavaField(TEMPLATE_DOCUMENT_FIELD,
                                        TemplateDOM.class.getName(),
                                        JavaModifiers.PRIVATE | JavaModifiers.STATIC | JavaModifiers.FINAL,
                                        "Template document shared by all instances.",
                                        null);
        docClass.addField(field);
        body.add(document.getClass().getName() + " " + DOCUMENT_VAR_NAME + " = ");
        domBuilderGenerator.createMethodCall(body);

        // Set up the template document.
        body.addln(TEMPLATE_DOCUMENT_FIELD + " = new " + TemplateDOM.class.getName()
                   + "(" + DOCUMENT_VAR_NAME + ");");

        if (fPreFormatter != null) {
            preFormatterInit(docClass);
        }
    }

    /**
     * Generate code to set preformat OutputOptions in document.
     */
    private void createPreformattedTextSetup(JavaCode body) {
        body.addln(LAZY_DOCUMENT_FIELD + ".setPreFormatOutputOptions("
                   + PREFORMAT_OUTPUT_OPTIONS_FIELD + ");");
    }

    /**
     * Create method that creates an unexpanded instance document.
     */
    private void createBuildDocument(XMLCDocument xmlcDoc,
                                     JavaClass docClass,
                                     JavaMethod buildDocumentMethod) {
        // Create document field
        JavaField docField = new JavaField(LAZY_DOCUMENT_FIELD,
                                           LazyDocument.class.getName(),
                                           JavaModifiers.PRIVATE,
                                           "Lazy DOM document", null);
        docClass.addField(docField);

        // Add code to build document
        JavaCode body = buildDocumentMethod.getCode();

        // Create from template using a XMLCDomFactory
        generateDocCreateFromTemplate(xmlcDoc.getDocument(),
                                      TEMPLATE_DOCUMENT_FIELD,
                                      LAZY_DOCUMENT_FIELD,
                                      body);
        if (fPreFormatter != null) {
            createPreformattedTextSetup(body);
        }

        // Save document information.
        body.addln("setDocument(" + LAZY_DOCUMENT_FIELD + ", "
                   + JavaLang.createStringConst(xmlcDoc.getDomFactory().getMIMEType()) + ", "
                   + JavaLang.createStringConst(xmlcDoc.getEncoding()) + ");");
    }

    /**
     * Generate the document builder method.
     * @see DocBuilderGenerator#createBuildDocumentMethod
     */
    public void createBuildDocumentMethod(XMLCDocument xmlcDoc,
                                          AccessorGenerator accessorGenerator,
                                          ElementTable elementTable,
                                          JavaClass docClass,
                                          JavaMethod buildDocumentMethod) throws XMLCException {
        fPreFormatter = new PreFormatter(xmlcDoc);
        createTemplateDocInit(xmlcDoc, accessorGenerator,
                              elementTable, docClass);
        createBuildDocument(xmlcDoc, docClass, buildDocumentMethod);
    }
}

