/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: LazyDOMNodeCreateGenerator.java,v 1.1.1.1 2003/03/10 16:36:22 taweili Exp $
 */

package org.enhydra.xml.xmlc.dom.lazydom;

import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.xmlc.codegen.JavaLang;
import org.enhydra.xml.xmlc.codegen.JavaCode;
import org.enhydra.xml.xmlc.dom.*;
import org.enhydra.xml.xmlc.dom.generic.NodeCreateGenerator;
import org.enhydra.xml.io.PreFormattedText;
import org.enhydra.xml.lazydom.*;
import org.w3c.dom.*;
import org.w3c.dom.html.HTMLDocument;
import org.enhydra.xml.xmlc.compiler.ClassGenerator;

/**
 * DOM node creation class that overrides methods in the generic version to
 * assign node ids.
 */
class LazyDOMNodeCreateGenerator extends NodeCreateGenerator {
    /**
     * If not null, the object to preformat node text.
     */
    private PreFormatter fPreFormatter;

    /**
     * Constructor.
     */
    public LazyDOMNodeCreateGenerator(XMLCDocument xmlcDoc,
                                      PreFormatter preFormatter) {
        super(xmlcDoc);
        fPreFormatter = preFormatter;
    }

    /**
     * Get the string constant for the preformatted text for a node, or "null"
     * if node should not be associated with the node.
     */
    private String getPreFormattedText(Node node) {
        String text = null;
        if ((node instanceof PreFormattedText) && (fPreFormatter != null)) {
            text = fPreFormatter.preFormatNode(node);
        }
        return JavaLang.createStringConst(text);
    }

    /**
     * Generate Document creation.
     */
    protected void genDocumentCreate(Document document,
                                     String docVar,
                                     String docTypeVar, 
                                     JavaCode body) {
        DocumentType docType = document.getDoctype();
        String docTypeRef;
        if (docType == null) {
            docTypeRef = "null";
        } else {
            docTypeRef = docTypeVar;
            genDocumentTypeCreate(docType, docVar, docTypeVar, body);
        }

        int nodeId = ((LazyNode)document).getNodeId();
        String docClass = document.getClass().getName();
        Element docElem = document.getDocumentElement();
        body.addln(docClass + " " + docVar + " = (" + docClass + ")"
                   + ClassGenerator.DOM_FACTORY_FIELD_NAME + ".createDocument("
                   + JavaLang.createStringConst(docElem.getNamespaceURI()) + ", "
                   + JavaLang.createStringConst(docElem.getNodeName()) + ", "
                   + docTypeRef + ");");
        body.addln(docVar + ".makeTemplateNode(" + nodeId + ");");
    }
    
    /**
     * Generate DocumentType creation.
     */
    protected void genDocumentTypeCreate(DocumentType docType,
                                         String docVarName,
                                         String varName, 
                                         JavaCode body) {
        String docTypeClass = docType.getClass().getName();
        int nodeId = ((LazyNode)docType).getNodeId();
        body.addln(docTypeClass + " " + varName
                   + " = (" + docTypeClass + ")" 
                   + ClassGenerator.DOM_FACTORY_FIELD_NAME
                   + ".createDocumentType("
                   + JavaLang.createStringConst(docType.getNodeName()) + ", "
                   + JavaLang.createStringConst(docType.getPublicId()) + ", " 
                   + JavaLang.createStringConst(docType.getSystemId()) + ", " 
                   + JavaLang.createStringConst(docType.getInternalSubset()) + ");");
        body.addln("((" + LazyNode.class.getName() + ")" + varName
                   + ").makeTemplateNode(" + nodeId + ");");
    }

    /**
     * Generate Entity creation.
     */
    protected void genEntityCreate(Entity entity,
                                   String docVarName,
                                   String varName, 
                                   JavaCode body) {
        int nodeId = ((LazyNode)entity).getNodeId();
        body.addln(varName + " = " + docVarName + ".createTemplateEntity("
                   + JavaLang.createStringConst(entity.getNodeName()) + ", "
                   + JavaLang.createStringConst(entity.getPublicId()) + ", "
                   + JavaLang.createStringConst(entity.getSystemId()) + ", "
                   + JavaLang.createStringConst(entity.getNotationName()) + ", "
                   + nodeId + ");");
    }

    /**
     * Generate Attr creation.
     */
    protected void genAttrCreate(Attr attr,
                                 String docVarName,
                                 String varName, 
                                 JavaCode body) {
        int nodeId = ((LazyNode)attr).getNodeId();
        body.add(varName + " = ");
        if (attr.getNamespaceURI() == null) {
            body.addln(docVarName + ".createTemplateAttribute("
                       + JavaLang.createStringConst(attr.getName()) + ", "
                       + nodeId + ");");
        } else {
            body.addln(docVarName + ".createTemplateAttributeNS("
                       + JavaLang.createStringConst(attr.getNamespaceURI()) + ", "
                       + JavaLang.createStringConst(attr.getName()) + ", "
                       + nodeId + ");");
        }
    }

    /**
     * Generate CDATASection creation.
     */
    protected void genCDATASectionCreate(CDATASection cdataSection,
                                         String docVarName,
                                         String varName, 
                                         JavaCode body) {
        int nodeId = ((LazyNode)cdataSection).getNodeId();
        body.addln(varName + " = " + docVarName
                   + ".createTemplateCDATASection("
                   + JavaLang.createStringConst(cdataSection.getData()) + ", "
                   + nodeId + ");");
    }

    /**
     * Generate Comment creation.
     */
    protected void genCommentCreate(Comment comment,
                                    String docVarName,
                                    String varName, 
                                    JavaCode body) {
        int nodeId = ((LazyNode)comment).getNodeId();
        body.addln(varName + " = " + docVarName + ".createTemplateComment("
                   + JavaLang.createStringConst(comment.getData()) + ", "
                   + nodeId + ");");
    }

    /**
     * Generate Element creation.
     */
    protected void genElementCreate(Element element,
                                    String docVarName,
                                    String varName,
                                    JavaCode body) {
        int nodeId = ((LazyNode)element).getNodeId();
        body.add(varName + " = ");
        if (isDocElement(element)) {
            String castVar = "((" + LazyElement.class.getName() + ")"
                + varName + ")";
            body.addln(docVarName + ".getDocumentElement();");
            body.addln(castVar + ".makeTemplateNode(" + nodeId + ");");
            body.addln(castVar + ".setPreFormattedText("
                       + getPreFormattedText(element) + ");");
        } else if (element.getNamespaceURI() == null) {
            body.addln(docVarName + ".createTemplateElement("
                       + JavaLang.createStringConst(element.getTagName()) + ", "
                       + nodeId + ", " + getPreFormattedText(element) + ");");
        } else {
            body.addln(docVarName + ".createTemplateElementNS("
                       + JavaLang.createStringConst(element.getNamespaceURI()) + ", "
                       + JavaLang.createStringConst(element.getTagName()) + ", "
                       + nodeId + ", " + getPreFormattedText(element) + ");");
        }
    }

    /**
     * Generate EntityReference creation.
     */
    protected void genEntityReferenceCreate(EntityReference entityRef,
                                            String docVarName,
                                            String varName, 
                                            JavaCode body) {
        int nodeId = ((LazyNode)entityRef).getNodeId();
        body.addln(varName + " = " + docVarName
                   + ".createTemplateEntityReference("
                   + JavaLang.createStringConst(entityRef.getNodeName()) + ", "
                   + nodeId + ");");
    }

    /**
     * Generate Notation creation.
     */
    protected void genNotationCreate(Notation notation,
                                     String docVarName,
                                     String varName, 
                                     JavaCode body) {
        int nodeId = ((LazyNode)notation).getNodeId();
        body.addln(varName + " = " + docVarName + ".createNotation("
                   + JavaLang.createStringConst(notation.getNodeName()) + ", "
                   + JavaLang.createStringConst(notation.getPublicId()) + ", "
                   + JavaLang.createStringConst(notation.getSystemId()) + ", "
                   + nodeId + ");");
    }

    /**
     * Generate ProcessingInstruction creation.
     */
    protected void genProcessInstructionCreate(ProcessingInstruction procInstr,
                                               String docVarName,
                                               String varName, 
                                               JavaCode body) {
        int nodeId = ((LazyNode)procInstr).getNodeId();
        body.addln(varName + " = " + docVarName
                   + ".createTemplateProcessingInstruction("
                   + JavaLang.createStringConst(procInstr.getNodeName()) + ", "
                   + JavaLang.createStringConst(procInstr.getData()) + ", "
                   + nodeId + ");");
    }

    /**
     * Generate Text creation.
     */
    protected void genTextCreate(Text text,
                                 String docVarName,
                                 String varName, 
                                 JavaCode body) {
        int nodeId = ((LazyNode)text).getNodeId();
        body.addln(varName + " = " + docVarName + ".createTemplateTextNode("
                   + JavaLang.createStringConst(text.getData()) + ", "
                   + nodeId + ", " + getPreFormattedText(text) + ");");
    }
}
