/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: SwingParser.java,v 1.2 2003/06/29 05:59:03 jkjome Exp $
 */

package org.enhydra.xml.xmlc.html.parsers.swing;

import org.enhydra.xml.xmlc.*;
import org.enhydra.xml.xmlc.misc.LineNumberMapReader;
import org.enhydra.xml.xmlc.misc.LineNumberMap;
import org.enhydra.xml.xmlc.compiler.*;
import org.enhydra.xml.xmlc.codegen.JavaLang;
import org.enhydra.xml.xmlc.html.*;
import org.enhydra.xml.xmlc.html.parsers.HTMLDocBuilder;
import org.enhydra.xml.xmlc.parsers.ParseTracer;
import org.enhydra.xml.xmlc.parsers.XMLCParser;
import org.enhydra.xml.xmlc.dom.*;
import org.enhydra.xml.io.ErrorReporter;
import org.w3c.dom.*;
import org.w3c.dom.Element;  // Work around compiler bug in above import.
import org.w3c.dom.html.*;
import java.util.*;
import java.net.*;
import java.io.*;
import javax.swing.text.BadLocationException;
import javax.swing.text.MutableAttributeSet;
import javax.swing.text.html.parser.DocumentParser;
import javax.swing.text.html.HTMLEditorKit;
import javax.swing.text.html.HTML;

/**
 * Interface to the Swing HTML parser.  This attempts to correct problem with
 * using the SWING HTML 3.2 parser to parse HTML 4.0 constructs and well as
 * general bugginess in the Swing parser.
 * <P>
 * The known Swing problems are:
 * <UL>
 * <LI> Swing 1.1 didn't allow LINK elements in the header.
 * <LI> Confusion around end forms (see processEndTagMisMatch),
 * </UL>
 */
class SwingParser extends HTMLEditorKit.ParserCallback {
    /**
     * Map constructed by reader of stream offset to filename
     * and line number.
     */
    private LineNumberMap fLineNumberMap;

    /**
     * Since the Swing parser saves an exception and continues to parse, 
     * we set this flag to stop the parsing on an error.
     */
    private boolean fGotError = false;

    /**
     * XMLC Document object.
     */
    private XMLCDocument fXmlcDoc;

    /**
     * The document builder.
     */
    private HTMLDocBuilder fDocBuilder;

    /**
     * Error thrown to stop parser.  This carries a causing exception
     * out of the parser.  This avoids the swing parser trying to handle
     * the error and creating cascading errors instead.
     */
    class ParserError extends Error {
        /**
         * Causing exception.
         */
        private Exception fCause;

        /**
         * Constructor.
         */
        public ParserError(Exception cause) {
            super(cause.getMessage());
            fCause = cause;
            fGotError = true;
        }

        /**
         * Get the cause.
         */
        public Throwable getCause() {
            return fCause;
        }
    }

    /**
     * Table of prefixes of SWING error messages to ignore.
     */
    private static final String[] fSwingIgnoredPrefixes = {
        "invalid.tagatt",          // Allow any attribute name.
        "javascript.unsupported",  // JavaScript should always be ignored :-)
        "req.att",
        "tag.unrecognized",
        "end.unrecognized",
        "unmatched.endtag",
        "multi.tagatt",
        "tag.ignoreimg"
    };
    
    /**
     * Ignore body tag errors for frame sets.
     */
    private static final String fIgnoreBodyPrefix = "tag.ignorebody";

    /**
     * Table of SWING unrecognized tags that we want to assume have and end
     * tag.
     */
     private static final HashSet fUnrecognizedAssumeEndTag = new HashSet();

    /**
     * Error reporter.
     */
    private ErrorReporter fReporter;

    /**
     * Verbose tracing.
     */
    private ParseTracer fTracer;

    /**
     * Is tracing enabled?
     */
    private boolean fTracingEnabled;

    /**
     * Static constructor.
     */
    static {
        fUnrecognizedAssumeEndTag.add("noscript");
        fUnrecognizedAssumeEndTag.add("frameset");
        fUnrecognizedAssumeEndTag.add("noframes");
        fUnrecognizedAssumeEndTag.add("span");
    }

    /**
     * Parse a file into a DOM.
     */
    void parse(HTMLDocBuilder docBuilder,
               ErrorReporter errorReporter,
               ParseTracer tracer,
               Reader input,
               LineNumberMap lineNumberMap) throws XMLCException, IOException {
        fDocBuilder = docBuilder;
        fXmlcDoc = docBuilder.getXMLCDocument();
        fLineNumberMap = lineNumberMap;
        fGotError = false;
        fReporter = errorReporter;
        fTracer = tracer;
        fTracingEnabled = ((fTracer != null) && fTracer.enabled());

        try {
            DocumentParser parser = new ParserAdaptor().getParser();
            parser.parse(input, this, true);
        } catch (ParserError err) {
            Throwable cause = err.getCause();
            if (cause instanceof XMLCException) {
                throw (XMLCException)cause;
            } else {
                throw new XMLCException(cause);
            }
        }
        Document document = fXmlcDoc.getDocument();
        if (fXmlcDoc.isHtmlFrameSet()) {
            cleanupFrameSet(document);
        }
        checkNodes(document);
    }

    /**
     * Generating an error message.
     */
    private void reportError(String msg, int pos) {
        LineNumberMap.Line rec = fLineNumberMap.getLineFromOffset(pos);
        fReporter.error(msg, rec.getFileName(), rec.getLineNum());
    }

    /**
     * Is an HTML.Tag object representing an end-tag,
     * as indicating by an ENDTAG attribute.
     */
     private boolean isEndTag(MutableAttributeSet attrSet) {
         return attrSet.isDefined(HTML.Attribute.ENDTAG);
     }

    /**
     * Determine if this one of the unknown tags that should be assumed
     * to have an end tag.
     */
    private boolean assumeEndTag(HTML.Tag tag) {
        return fUnrecognizedAssumeEndTag.contains(tag.toString());
    }

    /*
     * Output tag trace information.  Doesn't start a level of indentation.
     */
    private void traceTag(String message,
                          HTML.Tag tag,
                          MutableAttributeSet attrSet) {
        StringBuffer msg = new StringBuffer(message);
        msg.append(": ");
        msg.append(tag.toString());
        
        if (attrSet != null) {
            msg.append(":");
            Enumeration keyes = attrSet.getAttributeNames();
            while (keyes.hasMoreElements()) {
                Object attrKey = keyes.nextElement();
                Object attrValue = attrSet.getAttribute(attrKey);
                msg.append(" " + attrKey + "=\"" + attrValue + "\""); 
            }
        }
        fTracer.trace(msg.toString());
    }

    /*
     * Output start tag trace info.
     */
    private void traceStartTag(String message,
                               HTML.Tag tag,
                               MutableAttributeSet attrSet) {
        traceTag(message, tag, attrSet);
        fTracer.enter();
    }

    /*
     * Output end tag trace info.
     */
    private void traceEndTag(String message,
                             HTML.Tag tag) {
        StringBuffer msg = new StringBuffer(message);
        msg.append(": ");
        msg.append(tag.toString());
        fTracer.leave();
        fTracer.trace(msg.toString());
    }

    /**
     * Determine if an error should be ignored based on some crude
     * assumptions.
     */
    private boolean ignore(String errorMsg) {
        // Swing can't handle <body> in <noframes> section
        if (fXmlcDoc.isHtmlFrameSet()
            && errorMsg.startsWith(fIgnoreBodyPrefix)) {
            return true;
        }
        for (int i = 0; i < fSwingIgnoredPrefixes.length; i++) {
            if (errorMsg.startsWith(fSwingIgnoredPrefixes[i])) {
                return true;
            }
        }
        return false;
    }

    /**
     * Generate basic close-tag mismatch message.
     */
    private String tagMisMatchMsg(HTML.Tag tag) {
        return "Close tag mismatch, got </"
            + tag.toString().toUpperCase() + ">, expected </"
            + fDocBuilder.getCurrentNode().getNodeName() + ">";
    }

    /**
     * Handle end tag mismatch, trying to correct the problem.
     * Swing seems to get very confused by </FORM>,  generating
     * a end-tag event before all other tags are closed.
     */
    private void processEndTagMisMatch(HTML.Tag tag, int pos) {
        if (tag.toString().equalsIgnoreCase("form")
             || tag.toString().equalsIgnoreCase("span")) {
            // </FORM> when not expected.
            if (fTracingEnabled) {
                fTracer.trace("IGNORED: " + tagMisMatchMsg(tag)
                      + ", not modifying node stack");
            }
            return;
        }
        Node currentNode = fDocBuilder.getCurrentNode();
        if (currentNode.getNodeName().equalsIgnoreCase("form")
             || currentNode.getNodeName().equalsIgnoreCase("span")) {
            // <FORM> still on the stack, lets drop.
            if (fTracingEnabled) {
                fTracer.trace("IGNORED: " + tagMisMatchMsg(tag)
                      + ", appears to be left on stack from previous error; removing");
            }
            fDocBuilder.popCurrentNode();

            // Try again with next node on stack.
            handleEndTag(tag, pos);
            return;
        }

        // Don't have a clue whats going on, maybe a bad page.
        handleError(tagMisMatchMsg(tag), pos);
    }

    /**
     * Process a begin tag.
     */
    private void processBeginTag(HTML.Tag tag,
                                 MutableAttributeSet attrSet) {
        String tagName = tag.toString();

        fDocBuilder.startElement(tagName);
        if ((attrSet != null) && (attrSet.getAttributeCount() > 0)) {
            Enumeration keyes = attrSet.getAttributeNames();
            while (keyes.hasMoreElements()) {
                Object attrKey = keyes.nextElement();
                Object attrValue = attrSet.getAttribute(attrKey);
                if (attrValue != null) {
                    String name = attrKey.toString();
                    if (!name.equals(ParserAdaptor.IMPLIED_PSEUDO_ATTR)) {
                        fDocBuilder.addAttribute(name, attrValue.toString());
                    }
                }
            }
        }
    }

    /**
     * Process an end tag (real or synthetic).
     */
    private void processEndTag(HTML.Tag tag,
                               int pos) {
        Node currentNode = fDocBuilder.getCurrentNode();
        if ((currentNode != null)
            && !currentNode.getNodeName().equalsIgnoreCase(tag.toString())) {
            processEndTagMisMatch(tag, pos);
        } else {
            fDocBuilder.finishElement();
        }
    }

    /*
     * Flush method for interface; does nothing.
     * @see javax.swing.text.html.HTMLEditorKit.ParserCallback#flush
     */
    public void	flush() throws BadLocationException {
    }

    /**
     * Swing text handler.
     * @see javax.swing.text.html.HTMLEditorKit.ParserCallback#handleText
     */
    public void handleText(char[] data, int pos) {
        if (fGotError) {
            return;
        }
        if (fTracingEnabled) {
            fTracer.trace("TEXT: " + JavaLang.createStringConst(new String(data)));
        }
        fDocBuilder.addTextNode(new String(data));
    }

    /**
     * Swing comment handler.
     * @see javax.swing.text.html.HTMLEditorKit.ParserCallback#handleComment
     */
    public void handleComment(char[] data, int pos) {
        if (fGotError) {
            return;
        }
        if (fTracingEnabled) {
            fTracer.trace("COMMENT: " + JavaLang.createStringConst(new String(data)));
        }
        fDocBuilder.addComment(new String(data));
    }

    /**
     * Swing tag start handler.
     * @see javax.swing.text.html.HTMLEditorKit.ParserCallback#handleStartTag
     */
    public void handleStartTag(HTML.Tag tag, MutableAttributeSet attrSet, int pos) {
        if (fGotError) {
            return;
        }
        if (fTracingEnabled) {
            traceStartTag("BEGIN TAG", tag, attrSet);
        }
        processBeginTag(tag, attrSet);
    }

    /**
     * Swing end tag start handler.
     * @see javax.swing.text.html.HTMLEditorKit.ParserCallback#handleEndTag
     */
    public void handleEndTag(HTML.Tag tag, int pos) {
        if (fGotError) {
            return;
        }
        if (fTracingEnabled) {
            traceEndTag("END TAG", tag);
        }
        processEndTag(tag, pos);
    }

    /**
     * Swing simple tag start handler.
     * @see javax.swing.text.html.HTMLEditorKit.ParserCallback#handleSimpleTag
     */
    public void handleSimpleTag(HTML.Tag tag,
                                MutableAttributeSet attrSet,
                                int pos) {
        String tagName = tag.toString();

        // Swing 1.3 throws in special tag call at the end.
        if (fGotError || tagName.equals(ParserAdaptor.MAGIC_END_TAG)) {
            return;
        }
        if (fTracingEnabled) {
            traceTag("SIMPLE TAG", tag, attrSet);
        }

        /*
         * Unknown tags get special handling where we correct the tree when
         * the end tag is found.  This is necessary, as it was treated as a
         * simple tag when the opening tag was found.  SPAN also comes through
         * here in JDK 1.3.
         */
        if (isEndTag(attrSet)) {
            if ((tag instanceof HTML.UnknownTag) || tagName.equals("span")) {
                // unrecognized or span
                try {
                    fDocBuilder.fixUnrecognizedTagNesting(tagName);
                } catch (XMLCException except) {
                    reportError(except.toString(), pos);
                }
            }
        } else {
            processBeginTag(tag, attrSet);
            processEndTag(tag, pos);
        }
    }

    /**
     * Swing parser error handler that tries to manage the message
     * generated by parser that should really be ignored.
     * @see javax.swing.text.html.HTMLEditorKit.ParserCallback#handleError
     */
    public void handleError(String errorMsg,
                            int pos) {
        if (fGotError) {
            return;
        }
        boolean ignored = ignore(errorMsg);
        if (fTracingEnabled) {
            if (ignored) {
                fTracer.trace("IGNORE: " + errorMsg + ": " + pos);
            } else {
                fTracer.trace("ERROR: " + errorMsg + ": " + pos);
            }
        }
        if (!ignored) {
            reportError(errorMsg, pos);
        }
    }

    /*
     * Make sure a script element contains a comment.  This is 
     * necessary because the SWING parser drops the contents
     * if its not in comments.
     */
    private void checkScriptElement(HTMLScriptElement scriptElement) {
        if (scriptElement.getFirstChild() == null) {
            fReporter.error("Empty <SCRIPT> element found. The swing HTML parser discards javascript\n"
                            + " that is not in a comment");
        }
    }

    /**
     * Recursively perform somes checks on a node once complete.  This checks
     * for a script elements without comments, as text in such elements is
     * dropped.
     */
    private void checkNodes(Node node) {
        if (node instanceof HTMLScriptElement) {
            checkScriptElement((HTMLScriptElement)node);
        }
        for (Node child = node.getFirstChild(); child != null;
             child = child.getNextSibling()) {
            checkNodes(child);
        }
    }

    /**
     * Cleanup inserted top-level body in framesets that was added by Swing.
     */
    private void cleanupFrameSet(Document doc) {
        // Find <HTML> node.
        Node htmlElem = null;
        for (Node child = doc.getFirstChild(); child != null;
             child = child.getNextSibling()) {
            if (child instanceof HTMLHtmlElement) {
                htmlElem = child;
                break;
            }
        }
        if (htmlElem == null) {
            return;
        }
        
        // Find <BODY> node
        Node bodyElem = null;
        for (Node child = htmlElem.getFirstChild(); child != null;
             child = child.getNextSibling()) {
            if (child instanceof HTMLBodyElement) {
                bodyElem = child;
                break;
            }
        }
        if (bodyElem == null) {
            return;
        }

        // Move body children up one level.
        for (Node child = bodyElem.getFirstChild(); child != null;) {
            Node move = child;
            child = child.getNextSibling();

            bodyElem.removeChild(move);
            htmlElem.insertBefore(move, bodyElem);
        }
        
        // Finally, drop the <BODY>
        htmlElem.removeChild(bodyElem);
    }
}
