/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: TidyErrorHandler.java,v 1.1.1.1 2003/03/10 16:36:22 taweili Exp $
 */

package org.enhydra.xml.xmlc.html.parsers.tidy;

import org.enhydra.xml.xmlc.XMLCError;
import org.enhydra.xml.io.ErrorReporter;
import org.enhydra.xml.xmlc.misc.LineNumberMap;
import java.io.Writer;
import java.io.PrintWriter;
import java.io.IOException;
import gnu.regexp.*;


/**
 * Class for intercepting Tidy error output and doing custom error handing.
 * This is a PrintWrite that overrides all public methods to do our own
 * handling.  It parses output looking for errors an warnings and passes them
 * to the XMLC error handler.  This is hack, hopefully Tidy.parse will get
 * real error callbacks.  It will break if tidy is ever localized.
 */
class TidyErrorHandler extends PrintWriter {
    /**
     * Actual output stream.
     */
    private ErrorReporter fReporter;

    /**
     * Name of file being parsed.
     */
    private String fFileName;

    /**
     * If not null, we are using SSIReader and this is the mapping back to the
     * original line numbers.  Used to translate line numbers to source files.
     */
    private LineNumberMap fLineNumberMap;

    /**
     * Line buffer.  Use to hold message until end of line.
     */
    private StringBuffer fOutbuf = new StringBuffer();

    /**
     * Regular expression for parsing warning/error lines.
     * Result subexpressions are:
     *  1 - line number
     *  2 - "Warning", "Error" or "" (the empty string seems to be a tidy
     *      bug).
     *  3 - message
     * If these change, processWarnError must change.
     * 
     */
    private static final String WARN_ERROR_REGEXP = 
        "^line ([0-9,]+) column [0-9,]+ - (Warning:|Error:|) (.*)$";

    /**
     * Warning string value that is matched in WARN_ERROR_REGEXP.
     */
    private static final String WARN_MSG = "Warning:";

    /**
     * Compiler warning/error regexp.
     */
    private static final RE fWarnErrorRE;

    /**
     * Regular expressions defining lines to discard.
     */
    private static final String[] DISCARD_REGEXPS = {
        ".*Tidy \\(vers.*\\) Parsing.*",
        ".*\\\".*\\\" appears to be.*",
        ".*warnings/errors were found.*",
        ".*warnings or errors were found.*",
        ".*This document has errors that must be fixed before.*",
        ".*Document content looks like.*"
    };

    /**
     * Compiled expressions defining lines to discard.
     */
    private static final RE[] fDiscardREs;

    static {
        // Compile regular expressions.
        try {
            fWarnErrorRE = new RE(WARN_ERROR_REGEXP, 0,
                                  RESyntax.RE_SYNTAX_POSIX_EXTENDED);

            fDiscardREs = new RE[DISCARD_REGEXPS.length];
            for (int idx = 0; idx < DISCARD_REGEXPS.length; idx++) {
                fDiscardREs[idx] = new RE(DISCARD_REGEXPS[idx],
                                          RE.REG_MULTILINE,
                                          RESyntax.RE_SYNTAX_POSIX_EXTENDED);
            }
        } catch (REException except) {
            throw new XMLCError(except);
        }
    }

    /**
     * Constructor.
     */
    public TidyErrorHandler(ErrorReporter reporter,
                            String fileName,
                            LineNumberMap lineNumberMap) {
        // All print methods are overridden, so super is not actually used.
        super(System.err);
        fReporter = reporter;
        fFileName = fileName;
        fLineNumberMap = lineNumberMap;
    }

    /**
     * Convert string containing a Tidy integer, which might contain a `,'
     * to a number.
     */
    private int convertTidyInt(String intStr) {
        // Remove commas
        char[] strBuf = new char[intStr.length()];
        int cIdx = 0;
        for (int sIdx = 0; sIdx < intStr.length(); sIdx++) {
            char c = intStr.charAt(sIdx);
            if (c != ',') {
                strBuf[cIdx++] = c;
            }
        }
        return Integer.valueOf(new String(strBuf, 0 ,cIdx)).intValue();
    }

    /**
     * Process an warning or error match.
     */
    private void processWarnError(REMatch match) {
        String fileName = fFileName;
        int lineNum = convertTidyInt(match.toString(1));
        if (fLineNumberMap != null) {
            // Translate to source file and line
            LineNumberMap.Line line = fLineNumberMap.getLineFromLineNum(lineNum);
            fileName = line.getFileName();
            lineNum = line.getLineNum();
        }

        String which = match.toString(2);
        String msg = match.toString(3);

        // Treat lack of `Error' `Warn' as a warning
        if (which.equals(WARN_MSG) || (which.length() == 0)) {
            fReporter.warning(msg, fileName, lineNum);
        } else {
            fReporter.error(msg, fileName, lineNum);
        }
    }

    /**
     * Process completed line.
     */
    private void processLine() {
        REMatch match =  fWarnErrorRE.getMatch(fOutbuf);
        if (match != null) {
            processWarnError(match);
        } else {
            // See if this should be discarded
            int idx;
            for (idx = 0; idx < fDiscardREs.length; idx++) {
                if (fDiscardREs[idx].isMatch(fOutbuf)) {
                    break;
                }
            }
            if (idx == fDiscardREs.length) {
                // Didn't match a discard pattern, output it.
                fReporter.warning("XMLC did not detect this Tidy message: "
                                  + fOutbuf);
            }
        }
            
        fOutbuf.setLength(0);
    }

    /**
     * Flush the stream.
     */
    public void flush() {
    }

    /**
     * Close the stream.
     */
    public void close() {
    }

    /**
     * Flush the stream and check its error state.
     */
    public boolean checkError() {
        return false;
    }

    /**
     * Write a single character.
     */
    public void write(int c) {
        fOutbuf.append(c);
    }

    /**
     * Write a portion of an array of characters.
     */
    public void write(char buf[], int off, int len) {
        fOutbuf.append(buf, off, len);
    }

    /**
     * Write a portion of a string.
     */
    public void write(String s, int off, int len) {
        fOutbuf.append(s.substring(off, off+len));
    }

    /**
     * Write a string.
     */
    public void write(String s) {
        fOutbuf.append(s);
    }

    /**
     * Finish the line.
     */
    public void println() {
        processLine();
    }

    /**
     * Print a boolean, and then finish the line.
     */
    public void println(boolean x) {
        fOutbuf.append(x);
        processLine();
    }

    /**
     * Print a character, and then finish the line.
     */
    public void println(char x) {
        fOutbuf.append(x);
        processLine();
    }

    /**
     * Print an integer, and then finish the line.
     */
    public void println(int x) {
        fOutbuf.append(x);
        processLine();
    }

    /**
     * Print a long, and then finish the line.
     */
    public void println(long x) {
        fOutbuf.append(x);
        processLine();
    }

    /**
     * Print a float, and then finish the line.
     */
    public void println(float x) {
        fOutbuf.append(x);
        processLine();
    }

    /**
     * Print a double, and then finish the line.
     */
    public void println(double x) {
        fOutbuf.append(x);
        processLine();
    }

    /**
     * Print an array of characters, and then finish the line.
     */
    public void println(char x[]) {
        fOutbuf.append(x);
        processLine();
    }

    /**
     * Print a String, and then finish the line.
     */
    public void println(String x) {
        fOutbuf.append(x);
        processLine();
    }

    /**
     * Print an Object, and then finish the line.
     */
    public void println(Object x) {
        fOutbuf.append(x);
        processLine();
    }
}
