/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: LazyDOMBuilder.java,v 1.1.1.1 2003/03/10 16:36:26 taweili Exp $
 */

package org.enhydra.xml.lazydom;

import org.enhydra.xml.driver.TestException;
import org.enhydra.xml.xmlc.compiler.Parse;
import org.enhydra.xml.xmlc.metadata.MetaDataDocument;
import org.enhydra.xml.xmlc.metadata.MetaData;
import org.enhydra.xml.xmlc.metadata.InputDocument;
import org.enhydra.xml.xmlc.metadata.DocumentClass;
import org.enhydra.xml.xmlc.dom.XMLCDocument;
import org.enhydra.xml.xmlc.XMLObject;
import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.lazydom.LazyDocument;
import org.enhydra.xml.lazydom.TemplateDOM;
import org.enhydra.xml.io.ErrorReporter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.File;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentType;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * Class to handle creating LazyDOM document instances for tests, either by
 * creating or by parsing. Handles conversions from parsed LazyDOM to an
 * unexpanded LazyDOM.  This uses XMLC's wrapper around parser to handle the
 * parsing.  It would be nice if there XMLC parsing was more cleanly seperated
 * from the rest of XMLC.
 */
class LazyDOMBuilder {
    /** Compile-time debugging, enables more tracing */
    private static final boolean DEBUG = false;

    /** Signatures of constructors */
    private static final Class[] XML_CONSTR_SIGN
        = new Class[]{DocumentType.class, TemplateDOM.class};
    private static final Class[] HTML_CONSTR_SIGN
        = new Class[]{TemplateDOM.class};

    /** Trace output, non-null if enabled */
    private PrintWriter fTraceOut;
    
    /** Class for XMLC DOM factory */
    private Class fDomFactoryClass;

    /** Writer for messages. */
    private PrintWriter fMsgWriter;

    /** Constructor. */
    public LazyDOMBuilder(PrintWriter msgWriter,
                          Class domFactoryClass) {
        if (DEBUG) {
            fTraceOut = msgWriter;
        }
        fMsgWriter = msgWriter;
        fDomFactoryClass = domFactoryClass;
    }

    /** 
     * Parse a document using XMLC parser wrapper to do the dirty work.
     */
    private LazyDocument parseFile(File srcFile)
        throws IOException, XMLCException {
        Parse parser = new Parse(new ErrorReporter(fMsgWriter), fTraceOut);
        MetaData metaData = MetaDataDocument.newInstance().getMetaData();
        InputDocument inputDoc = metaData.getInputDocument();
        inputDoc.setUrl(srcFile.getPath());
        DocumentClass docClass = metaData.getDocumentClass();
        docClass.setDomFactory(fDomFactoryClass.getName());

        // HTML test pages have some warnings, just drop em
        metaData.getParser().setWarnings(false);
        
        if (DEBUG) {
            metaData.getCompileOptions().setVerbose(true);
            metaData.getCompileOptions().setPrintDocumentInfo(true);
        }

        XMLCDocument xmlcDoc = parser.parse(metaData);
        return (LazyDocument)xmlcDoc.getDocument();
    }

    /**
     * Attempt to createa a new instance of a XML LazyDocument, given a
     * template.
     */
    private LazyDocument newXmlLazyDoc(Class lazyDocumentClass,
                                       TemplateDOM templateDOM)
        throws NoSuchMethodException, InstantiationException,
               IllegalAccessException, InvocationTargetException {
        Constructor constr
            = lazyDocumentClass.getConstructor(XML_CONSTR_SIGN);
        return (LazyDocument)constr.newInstance(new Object[]{null, templateDOM});
    }

    /**
     * Attempt to createa a new instance of a HTML LazyDocument, given a
     * template.
     */
    private LazyDocument newHtmlLazyDoc(Class lazyDocumentClass,
                                       TemplateDOM templateDOM)
        throws NoSuchMethodException, InstantiationException,
               IllegalAccessException, InvocationTargetException {
        Constructor constr
            = lazyDocumentClass.getConstructor(HTML_CONSTR_SIGN);
        return (LazyDocument)constr.newInstance(new Object[]{templateDOM});
    }

    /**
     * Create a new instance of a LazyDocument, given a template.
     */
    private LazyDocument newLazyDoc(Class lazyDocumentClass,
                                    TemplateDOM templateDOM) {
        // Try for XML, then HTML constructor pattern
        try {
            try {
                return newXmlLazyDoc(lazyDocumentClass, templateDOM);
            } catch (NoSuchMethodException except) {
                // ignored
            }
            return  newHtmlLazyDoc(lazyDocumentClass, templateDOM);
        } catch (Exception except) {
            throw new TestException("failed to create instance of: " + lazyDocumentClass,
                                    except);
        }
    }

    /**
     * Parse a document into an expanded LazyDOM.
     */
    public LazyDocument parseUnexpanded(File srcFile) {
        try {
            LazyDocument expandedDoc = parseFile(srcFile);
            TemplateDOM templateDom = new TemplateDOM(expandedDoc);
            return newLazyDoc(expandedDoc.getClass(), templateDom);
        } catch (IOException except) {
            throw new TestException(srcFile.getPath(), except);
        } catch (XMLCException except) {
            throw new TestException(srcFile.getPath(), except);
        }
    }
            
}


