/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: OasisTestDriver.java,v 1.1.1.1 2003/03/10 16:36:27 taweili Exp $
 */

package org.enhydra.xml.oasis;

import java.io.*;
import org.w3c.dom.*;
import org.enhydra.apache.xerces.parsers.DOMParser;
import org.xml.sax.SAXException;
import org.xml.sax.InputSource;

/**
 * Test driver class for Oasis XML conformance tests.  This parses the test
 * descriptions and calls derived test methods for each test.
 */
public class OasisTestDriver {
    /*
     * Default location of file describing the tests.
     */
    public static final String XML_CONF = "conf/xmlconf.xml";

    /**
     * Test conf file to use.
     */
    private String fTestConf;

    /**
     * Flag used to determine if a tests was executed.
     */
    private boolean fRanUserTest;

    /**
     * Count of sucessful tests.
     */
    private int fSuccessCount = 0;

    /**
     * Count of skipped tests.
     */
    private int fSkippedCount = 0;

    /**
     * Count of disabled tests.
     */
    private int fDisabledTestCount = 0;

    /**
     * Count of disabled test casess.
     */
    private int fDisabledTestCaseCount = 0;

    /**
     * Count of failed tests.
     */
    private int fFailedCount = 0;

    /**
     * Constructor.
     * @param testConf Test conf XML file to use or null for default.
     */
    protected OasisTestDriver(String testConf) {
        fTestConf = (testConf != null) ? testConf : XML_CONF;
    }

    /**
     * Get count of sucessful tests.
     */
    public int getSuccessCount() {
        return fSuccessCount;
    }
    
    /**
     * Get count of skipped tests (no derived handler).
     */
    public int getSkippedCount() {
        return fSkippedCount;
    }
    
    /**
     * Get count of disabled tests.
     */
    public int getDisabledTestCount() {
        return fDisabledTestCount;
    }
    
    /**
     * Get count of disabled test cases.
     */
    public int getDisabledTestCaseCount() {
        return fDisabledTestCaseCount;
    }
    
    /**
     * Get count of failed tests.
     */
    public int getFailedCount() {
        return fFailedCount;
    }
    
    /**
     * Method to run a type `valid' test.  Override to execute these tests.
     */
    protected void runValidTest(Test test) throws Exception {
        fRanUserTest = false;  // User didn't override
    }

    /**
     * Method to run a type `invalid' test.  Override to execute these tests.
     */
    protected void runInvalidTest(Test test) throws Exception {
        fRanUserTest = false;  // User didn't override
    }

    /**
     * Method to run a type `not-wf' test.  Override to execute these tests.
     */
    protected void runNotWFTest(Test test) throws Exception {
        fRanUserTest = false;  // User didn't override
    }

    /**
     * Method to run a type `error' test.  Override to execute these tests.
     */
    protected void runErrorTest(Test test) throws Exception {
        fRanUserTest = false;  // User didn't override
    }

    /**
     * Record a test success.  Outputs to stderr unless overridden.
     */
    protected void reportSuccess(Test test) {
        System.err.println("Passed: " + test.getId());
    }

    /**
     * Record a test failure.  Outputs to stderr unless overridden.
     */
    protected void reportFailure(Test test,
                                 Throwable failure) {
        System.err.println("*** Failed: " + test.getId()
                           + ": " + test.getURI()
                           + ": " + failure.toString());
        failure.printStackTrace();
    }

    /**
     * Run a test and record the results.
     */
    private void runTest(Test test) {
        // Flag to determine if the user overrode the test method.
        fRanUserTest = true;

        try {
            if (test.getType().equals(Test.TYPE_VALID)) {
                runValidTest(test);
            } else if (test.getType().equals(Test.TYPE_INVALID)) {
                runInvalidTest(test);
            } else if (test.getType().equals(Test.TYPE_NOT_WF)) {
                runNotWFTest(test);
            } else if (test.getType().equals(Test.TYPE_ERROR)) {
                runErrorTest(test);
            }
            if (fRanUserTest) {
                fSuccessCount++;
                reportSuccess(test);
            } else {
                fSkippedCount++;
            }
        } catch (Throwable failure) {
            fFailedCount++;
            reportFailure(test, failure);
        }
    }

    /**
     * Recursivly run tests.
     */
    private void runTestCases(TestCases testCases) {
        Test[] childTests = testCases.getTests();
        for (int idx = 0; idx < childTests.length; idx++) {
            if (childTests[idx].isDisabled()) {
                fDisabledTestCount++;
            } else {
                runTest(childTests[idx]);
            }
        }
        
        TestCases[] childTestCases = testCases.getTestCases();
        for (int idx = 0; idx < childTestCases.length; idx++) {
            if (childTestCases[idx].isDisabled()) {
                fDisabledTestCaseCount++;
            } else {
                runTestCases(childTestCases[idx]);
            }
        }
    }

    /**
     * Run tests.
     */
    protected void runTests() throws IOException, SAXException {
        TestSuite testSuite = TestSuite.parse(fTestConf);
        runTestCases(testSuite);
    }
}

