/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: Test.java,v 1.1.1.1 2003/03/10 16:36:27 taweili Exp $
 */

package org.enhydra.xml.oasis;

import java.util.ArrayList;
import java.io.PrintWriter;
import java.io.File;
import org.xml.sax.Attributes;

/**
 * Definition of a test.
 */
public final class Test {
    /** Constants for the type of (external) ENTITIES required */
    public static final String ENTITIES_BOTH = "both";
    public static final String ENTITIES_NONE = "none";
    public static final String ENTITIES_PARAMETER = "parameter";
    public static final String ENTITIES_GENERAL = "general";

    /** Constants for the type of the test */
    public static final String TYPE_VALID = "valid";
    public static final String TYPE_INVALID = "invalid";
    public static final String TYPE_NOT_WF = "not-wf";
    public static final String TYPE_ERROR = "error";

    /** Parent for test */
    private TestCases fParent;

    /** Directory containing tests */
    private File fTestDir;

    /** Natual-language description of the test */
    private String fDescription = "";

    /** The type of (external) ENTITIES required */
    private String fEntities;

    /** The unique test id */
    private String fId;

    /** The file containing the expected canonical output */
    private String fOutput;

    /** The file containing the expected third canonical output */
    private String fOutput3;

    /** The sections of the XMLC spec that are tested */
    private String fSections;

    /** The type of the test. */
    private String fType;

    /** URI of the test document */
    private String fURI;

    /** Is this test disabled? */
    private boolean fDisabled;

    /** Reasons for the test being disabled */
    private String fDisabledReason = "";

    /**
     * Constructor.
     */
    protected Test(TestCases parent,
                   File testDir,
                   Attributes attrs) {
        fParent = parent;
        fTestDir = testDir;
        fEntities = TestSuiteParser.getAttr(attrs, "ENTITIES");
        fId = TestSuiteParser.getAttr(attrs, "ID");
        fOutput = TestSuiteParser.getAttr(attrs, "OUTPUT");
        fOutput3 = TestSuiteParser.getAttr(attrs, "OUTPUT3");
        fSections = TestSuiteParser.getAttr(attrs, "SECTIONS");
        fType = TestSuiteParser.getAttr(attrs, "TYPE");
        fURI = TestSuiteParser.getAttr(attrs, "URI");
    }
    
    /**
     * Get the parent, or null if none.
     */
    public TestCases getParent() {
        return fParent;
    }

    /**
     * Append characters to the description.
     */
    protected void appendDescription(String desc) {
        fDescription += desc;
    }

    /**
     * Append characters to the disabled reason.
     */
    protected void appendDisableReason(String reason) {
        fDisabledReason += reason;
    }

    /**
     * Flag as disabled.
     */
    protected void setDisabled() {
        fDisabled = true;
    }

    /**
     * Get the natural-language description of the test.
     */
    public String getDescription() {
        return fDescription;
    }

    /** 
     * Get the type of (external) ENTITIES required.
     */
    public String getEntities() {
        return fEntities;
    }

    /** 
     * Get the unique test id.
     */
    public String getId() {
        return fId;
    }

    /**
     * Get the file containing the expected canonical output.
     */
    public String getOutput() {
        return fOutput;
    } 

    /** 
     * Get the file containing the expected third canonical output.
     */
    public String output3() {
        return fOutput3;
    }

    /**
     * The sections of the XMLC spec that are tested.
     */
    public String getSections() {
        return fSections;
    }

    /**
     * The type of the test.
     */
    public String getType() {
        return fType;
    }

    /**
     * Get the URI of the test document.  This is relative to
     * the xmlconf.xml.
     */
    public String getURI() {
        return fURI;
    }

    /**
     * Get the full URI of the test document.
     */
    public File getTestFile() {
        return new File(fTestDir, fURI);
    }

    /**
     * Is the test disabled.
     */
    public boolean isDisabled() {
        return fDisabled;
    }

    /**
     * Get the reason the test was disabled.
     */
    public String getDisabledReason() {
        return fDisabledReason;
    }

    /**
     * Print the contents.
     */
    public void print(PrintWriter writer,
                      int indent) {
        Print.printlnIndented(writer, indent, "Test: "
                              + fId + ": " + fURI);
    }
}
