/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: XmlcBasicTest.java,v 1.1.1.1 2003/03/10 16:36:27 taweili Exp $
 */

package org.enhydra.xml.xmlc.driver;

import java.io.PrintWriter;
import java.io.File;
import java.util.ArrayList;
import org.enhydra.xml.driver.TestFileOps;
import org.enhydra.xml.driver.TestException;
import org.enhydra.xml.driver.TestError;
import org.enhydra.xml.xmlc.driver.ExecXmlc;
import org.enhydra.xml.xmlc.driver.OutputDocument;
import org.enhydra.xml.xmlc.driver.XmlcTestParams;
import org.enhydra.xml.xmlc.XMLObject;
import org.enhydra.xml.xmlc.XMLCException;
import org.enhydra.xml.io.OutputOptions;

/**
 * Class to run a basic set of XMLC tests given a set of parameters.
 * This class is extended to create HTML- and XML-specified basic
 * test, and those are extended to implement specific tests.
 * 
 * <UL>
 * <LI> Compile a page with the specified set of parameters.
 * <LI> Compare output of compile dump of DOM to expected
 * <LI> Load the compiled document class.
 * <LI> Compare dump of compiled DOM to expected
 * <LI> Output the document.
 * <LI> Compiled the new document.
 * <LI> Compare output of compile dump of DOM to expected.
 * </UL>
 *
 * The following general parameters are use to reuse the specific tests
 * with different overall configurations of XMLC.  These define the tests
 * ids and output classes used.
 * <UL>
 * <LI> parser - The HTML parser to use: tidy or swing.
 * <LI> dom - The HTML DOM implementation to use.
 * <LI> autoload - Should auto compilation/loading testing be done?
 * </UL>
 */
public class XmlcBasicTest {
    /** Test case we are assocaited with. */
    private XmlcTestCaseBase fTest;

    /** Source file being compiled */
    private File fSrcFile;
    
    /**
     * Extra source files (DTDs, etc) that need to be copied to the
     * same directory as the source file for second compile pass or
     * for reloading tests.  Null, if none are specified, 
     */
    private ArrayList fExtraSrc;

    /** Doc type extension (e.g. html, xml) */
    private String fDocFileExt;

    /** Is the compiled step expected to fail? */
    private boolean fCompileShouldFail;

    /** Option options to use instead of defaults */
    private OutputOptions fOutputOptions;

    /** Write tracing if not null */
    private PrintWriter fVerboseOut;

    /** Compile object for first compile of document */
    private ExecXmlc fComp1Xmlc;

    /** Compile object for compile of generated document */
    private ExecXmlc fComp2Xmlc;

    /** Encoding that should in in the document */
    private String fExpectedEncoding;

    /** Should toDocument() be used to write the object */
    private boolean fUseToDoc;

    /** Should the reloading tests edit the source or just touch it? */
    private boolean fEditReloadSource = true;

    /** Flag indicated if expected files should be diffed or saved */
    private boolean fDiffExpected = true;

    /** 
     * Constructor.
     * @param unqualTestId Id for test
     */
    public XmlcBasicTest(XmlcTestCaseBase test,
                         String docFileExt,
                         File srcFile) {
        fTest = test;
        fDocFileExt = docFileExt;
        fSrcFile = srcFile;
        fComp1Xmlc = new ExecXmlc(test);
        fComp1Xmlc.setSrcFile(fSrcFile);

        fComp2Xmlc = new ExecXmlc(test);
    }

    /** Get the test case we are associated with */
    public XmlcTestCaseBase getTestCase() {
        return fTest;
    }

    /**
     * Used to indicate if expected files should be diffed and saved.  The
     * default is yes, but for some tests, the results are so large that it's
     * not worth the overhead of keeping them.
     */
    public void disableExpectedFileDiff() {
        fDiffExpected = false;
    }

    /**
     * Indicate if the edited source should be edited or touched.  In some
     * cases (e.g. a document of all entities), the editing doesn't actually
     * change anything, causing a check to fail.  So we don't edit or do the
     * check, just touch.
     */
    public void setEditReloadSource(boolean val) {
        fEditReloadSource = val;
    }

    /**
     * Add an associated file that need to be in the same directory as
     * the document to recompile it.  These are files such as DTDs, SSIs,
     * etc.  These are also used for reloading tests.
     */
    public void addExtraSrcFile(File extraSrcFile) {
        if (fExtraSrc == null) {
            fExtraSrc = new ArrayList();
        }
        fExtraSrc.add(extraSrcFile);
        if (fTest.getParams().getReloading()) {
            fComp1Xmlc.addExtraSrcFile(extraSrcFile);
        }
    }

    /** 
     * Set the encoding that should in in the document
     */
    public void setExpectedEncoding(String encoding) {
        fExpectedEncoding = encoding;
    }

    /** 
     * Get the output options associate with the object.  If none exist,
     * they are created.
     */
    public OutputOptions getOutputOptions() {
        if (fOutputOptions == null) {
            fOutputOptions = new OutputOptions();
        }
        return fOutputOptions;
    }

    /** Indicate that toDocument() should used to write the object */
    public void setUseToDocument() {
        fUseToDoc = true;
    }

    /** Add an option to the first compile pass */
    public void addFirst(String opt) {
        fComp1Xmlc.addOpt(opt);
    }

    /** Add and option and argument to the first compile pass */
    public void addFirst(String opt,
                         String arg) {
        fComp1Xmlc.addOpt(opt, arg);
    }

    /** Add an option and two arguments to the first compile pass */
    public void addFirst(String opt,
                         String arg1,
                         String arg2) {
        fComp1Xmlc.addOpt(opt, arg1, arg2);
    }

    /** Add an option to the both compile passes */
    public void addBoth(String opt) {
        fComp1Xmlc.addOpt(opt);
        if (fComp2Xmlc != null) {
            fComp2Xmlc.addOpt(opt);
        }
    }

    /** Add and option and argument to the both compile passes */
    public void addBoth(String opt,
                        String arg) {
        fComp1Xmlc.addOpt(opt, arg);
        if (fComp2Xmlc != null) {
            fComp2Xmlc.addOpt(opt, arg);
        }
    }

    /** Add and option and argument to the both compile passes */
    public void addBoth(String opt,
                        File arg) {
        fComp1Xmlc.addOpt(opt, arg);
        if (fComp2Xmlc != null) {
            fComp2Xmlc.addOpt(opt, arg);
        }
    }

    /** Add an option and two arguments to the both compile passes */
    public void addBoth(String opt,
                        String arg1,
                        String arg2) {
        fComp1Xmlc.addOpt(opt, arg1, arg2);
        if (fComp2Xmlc != null) {
            fComp2Xmlc.addOpt(opt, arg1, arg2);
        }
    }

    /** Specify the -dom or -domfactory option, preventing automatic setting */
    public void setDomOpt(String opt,
                          String arg) {
        fComp1Xmlc.setDomOpt(opt, arg);
        fComp2Xmlc.setDomOpt(opt, arg);
    }

    /**
     * Add an XMLC option or metadata file.
     */
    public void addOptionFile(File file) {
        fComp1Xmlc.addOptionFile(file);
        if (fComp2Xmlc != null) {
            fComp2Xmlc.addOptionFile(file);
        }
    }

    /**
     * Set the compile-should-fail flag.
     */
    public void setCompileShouldFail() {
        fCompileShouldFail = true;
    }

    /**
     * Build a ExecXmlc object with argument required for the initial compile
     * and the recompile/dump.  This is implemented by the derived class.
     */
    protected void addStdOpts(ExecXmlc execXmlc) {
    }

    /**
     * Handle the compile failing
     */
    private void handleCompileFail(TestException testExcept) {
        if (fCompileShouldFail) {
            Throwable cause = testExcept.getCause();
            // Failure was expected, make sure cause is not a another test
            // exception or has not cause
            if ((cause == null) || (cause instanceof TestException)
                || (cause instanceof TestError)) {
                throw new TestException("compile that should fail, failed in test code",
                                        testExcept);
            }
        } else {
            // Should not have failed
            throw testExcept;
        }
    }


    /**
     * Do initial compile, creating log file and dump file
     */
    public void compileOriginalStep() {
        fTest.getGenClassFile().delete();
        addStdOpts(fComp1Xmlc);
        if (fTest.getParams().getReloading()) {
            fComp1Xmlc.addOpt(ExecXmlc.OPT_FOR_RECOMP);
        }
        fComp1Xmlc.addOpt(ExecXmlc.OPT_DEST_DIR, fTest.getClassRoot());
        fComp1Xmlc.addOpt(ExecXmlc.OPT_CLASS, fTest.getTestClass());
        fComp1Xmlc.addOpt(ExecXmlc.OPT_METHODS);
        fComp1Xmlc.addOpt(ExecXmlc.OPT_DUMP);

        // Compile with checking for expected failure
        File result = fTest.getResultFile(XmlcTestCaseBase.COMPILE_OUTPUT_EXT);
        boolean failed = false;
        try {
            TestFileOps.ensureFileDir(result);
            fComp1Xmlc.compile(result);
        } catch (TestException testExcept) {
            failed = true;
            handleCompileFail(testExcept);
        }
        if (fCompileShouldFail && !failed) {
            // Should have failed but didn't
            throw new TestException("compile should have failed");
        }
        if (fDiffExpected) {
            File expected = fTest.getLoadInvarExpectedFile(XmlcTestCaseBase.COMPILE_OUTPUT_EXT);
            fTest.getDiffer().diff(expected, result);
        }
    }

    /* 
     * Get the path to the file that was generated from outputing the compiled
     * document.
     */
    public File getGenerateFile() {
        return fTest.getResultFile("gen." + fDocFileExt);
    }

    /** Get path to file generated by the -docout on the second compile */
    public File getRegenerateFile() {
        return fTest.getResultFile("regen." + fDocFileExt);
    }

    /**
     * Output the compiled document.  
     */
    private void outputDocument(XMLObject doc) {
        if ((fOutputOptions != null) && fUseToDoc) {
            throw new TestError("bug: both output options and toDocument() are specified");
        }
        if (fUseToDoc) {
            OutputDocument.writeWithToDoc(doc, getGenerateFile());
        } else {
            // Uses default options if output options was not set.
            OutputDocument.write(doc, getGenerateFile(), fOutputOptions);
        }
    }

    /**
     * Parse the generated file using xmlc, dumping the DOM and other
     * information.  Since the generate files are written into directories
     * that are dependent on the DOM and parser, they can end up with
     * location-specific paths in the compiler output, causing failures on
     * diff.  To compensate for this, we filter the files with regular 
     * expressions first.
     */
    private void parseGenerated() {
        addStdOpts(fComp2Xmlc);
        fComp2Xmlc.addOpt(ExecXmlc.OPT_NO_COMPILE);
        fComp2Xmlc.addOpt(ExecXmlc.OPT_METHODS);
        fComp2Xmlc.addOpt(ExecXmlc.OPT_DUMP);
        fComp2Xmlc.addOpt(ExecXmlc.OPT_DOC_OUT, getRegenerateFile().getPath());
        fComp2Xmlc.setSrcFile(getGenerateFile());

        File out = fTest.getResultFile(XmlcTestCaseBase.PARSE_GEN_OUTPUT_EXT);
        fComp2Xmlc.compile(out);

        if (fDiffExpected) {
            File expected = fTest.getExpectedFile(XmlcTestCaseBase.PARSE_GEN_OUTPUT_EXT);
            fTest.getDiffer().diff(expected, out);
        }
    }

    /** 
     * Output the generated class and then parse the source.
     */
    public void reparseStep(XMLObject doc) {
        // output and verify DOM first..
        if (fDiffExpected) {
            fTest.dumpVerifyDom(doc, XmlcTestCaseBase.LOAD_DOM_EXT);
        } else {
            fTest.dumpDom(doc, XmlcTestCaseBase.LOAD_DOM_EXT);
        }

        if ((fExpectedEncoding != null)
            && !fExpectedEncoding.equals(doc.getEncoding())) {
            throw new TestException("expected document encoding of \""
                                    + fExpectedEncoding +" \", got \""
                                    + doc.getEncoding() + "\"");
        }

        if (fExtraSrc != null) {
            TestFileOps.copyFilesToDir(fExtraSrc, fTest.getResultsDir());
        }
        outputDocument(doc);
        parseGenerated();
    }

    /** 
     * Output the genreated class and then parse the source.
     */
    public void reparseStep() {
        reparseStep(fTest.loadTestDocument(fTest.getTestClass()));
    }

    /**
     * Execute the standard `basic test'.
     */
    public void basicTest() {
        compileOriginalStep();
        if (fCompileShouldFail) {
            return;  // should have failed
        }
	reparseStep();
    }
}
