/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: XmlcTestCaseBase.java,v 1.1.1.1 2003/03/10 16:36:27 taweili Exp $
 */

package org.enhydra.xml.xmlc.driver;

import java.io.PrintWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import org.enhydra.xml.driver.TestCaseBase;
import org.enhydra.xml.driver.TestError;
import org.enhydra.xml.driver.TestFileOps;
import org.enhydra.xml.driver.TestException;
import org.enhydra.xml.xmlc.XMLObject;
import org.enhydra.xml.xmlc.XMLCFactory;
import org.enhydra.xml.xmlc.XMLCLogger;
import org.enhydra.xml.xmlc.XMLCStdFactory;
import org.enhydra.xml.xmlc.deferredparsing.XMLCDeferredParsingFactory;
import org.enhydra.xml.dom.DOMInfo;
import org.enhydra.xml.dom.DOMStats;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * Base class for XMLC test case classes.  The <CODE>suite()<CODE> method
 * creates instances of this test case for all combinations of the params
 * object.
 */
public class XmlcTestCaseBase extends TestCaseBase {
    //FIXME: test an alternative naming
    public static final boolean ALT_NAMES = true;

    /** Standard result file extensions */
    public static final String COMPILE_OUTPUT_EXT = "compile.out";
    public static final String LOAD_DOM_EXT = "load.dom";
    public static final String PARSE_GEN_OUTPUT_EXT = "parsegen.out";
    public static final String PARSE_DOM_DUMP_EXT = "parse.dom";
    public static final String DOC_DOM_DUMP_EXT = "doc.dom";

    /** Empty arrays used to call reflection */
    private static final Class[] NO_PARAMS = new Class[0];
    private static final Class[] STR_PARAM = new Class[]{
        String.class
    };
    private static final Object[] NO_ARGS = new Object[0];

    /** Object containing test parameters */
    protected final XmlcTestParams fParams;

    /** Source output directory when keeping generate file   */
    private static final String SOURCE_DIR = "src";

    /** 
     * Indicates if an implementaion/interface is being used instead of 
     * class.  Set automatically is reloading testing, otherwise must
     * be set explictly
     */
    private boolean fUsingImplInterface;

    /** Factory for creating document objects. */
    private XMLCFactory fXMLCFactory;

    /** Test package name derived from parameters (lazy) */
    private String fTestPackage;

    /** Reload invarient expected results result dir (lazy) */
    private File fLoadInvarExpectedDir;

    /** Make a test id that includes that parameters */
    private static String mkTestId(Method method,
                                   XmlcTestParams params) {
        if (ALT_NAMES) {
            return method.getName() + "." + params.getParamsStr();
        } else {
            return params.getParamsStr() + "." + method.getName();
        }
    }

    /** Constructor. */
    public XmlcTestCaseBase(Method method,
                            XmlcTestParams params) {
        super(mkTestId(method, params), method);
        fParams = params;
        if (fParams.getReloading()) {
            setUsingImplInterface();
        }
    }

    /**
     * Get the parameters.
     */
    public XmlcTestParams getParams() {
        return fParams;
    }

    /** Set the XMLC factory to use for test */
    public void setFactory(XMLCFactory factory) {
        fXMLCFactory = factory;
    }

    /** Set the XMLC factory to use for test */
    private XMLCFactory createFactory(ClassLoader loader, XMLCLogger logger) {
	XMLCFactory factory;

        if (fParams.getIsDeferredParsing()) {
	    factory = new XMLCDeferredParsingFactory (null, loader, logger); // use reloading loader
	} else {
	    factory = new XMLCStdFactory(loader, logger); // use standard loader
	}

	return factory;
    }

    /** Indicate that an implementation/interface is being used. */
    public void setUsingImplInterface() {
        fUsingImplInterface = true;
        if (fXMLCFactory == null) {
            // Must go through factory to get object
	    fXMLCFactory = createFactory(getTestClassLoader(), null);
        }
    }

    /**
     * Get the test package, this merges in the test parameters to create
     * a package uniqu to this parameterization.  This override changes
     * the value of all of the other directories.
     */
    public String getTestPackage() {
        if (fTestPackage == null) {
            if (ALT_NAMES) {
                // Need to add `Pkg' to prevent conflict with class name
                fTestPackage = getClass().getName() + "Pkg" + "."
                    + fParams.getParamsStr();
                
            } else {
            // Split test case class name, inserting paramter info before
            // unqual class name.  This good parameterizations together.
            String clName = getClass().getName();
            int dotIdx = clName.lastIndexOf('.');
            String clFirst = clName.substring(0, dotIdx);
            String clLast = clName.substring(dotIdx+1);

            fTestPackage = clFirst + "." + fParams.getParamsStr() + "."
                + clLast;
            }
        }
        return fTestPackage;
    }
    
    /** 
     * Get the test document class name.  This is the name to specify to the
     * compiler, it maybe the name of the generated class or interface
     * dependening on the compile options.
     */
    public String getTestClass() {
        return getTestPackage() + "." + getTestName();
    }

    /**
     * Get the test document implementation.  This is the Impl name
     * if seperated interfaces and implementations are being generated,
     * otherwise it is the plain class name.
     */
    public String getTestClassImpl() {
        if (fUsingImplInterface) {
            return getTestClass() + "Impl";
        } else {
            return getTestClass();
        }
    }

    /**
     * Get the test document interface, if seperate interface and
     * implementation is being generated.
     */
    public String getTestClassInterface() {
        if (!fUsingImplInterface) {
            throw new TestError("request for interface when one is not being used");
        }
        return getTestClass();
    }

    /** 
     * Get the test expected results directory for reloading-invarient
     * expected files.
     */
    public File getLoadInvarExpectedDir() {
        if (fLoadInvarExpectedDir == null) {
            // Remove last directory entry
            fLoadInvarExpectedDir
                = new File(getExpectedRoot(),
                           getRelResultsDir().getParentFile().getPath());
        }
        return fLoadInvarExpectedDir;
    }
    
    /** Get reload-invarient expected file name. */
    public File getLoadInvarExpectedFile(String fname,
                                         String suffix) {
        return new File(getLoadInvarExpectedDir(),
                        fname + "." + suffix); 
    }

    /** Get an expected file name. */
    public File getLoadInvarExpectedFile(String suffix) {
        return getLoadInvarExpectedFile(getTestName(), suffix); 
    }


    /** Get source output directory using when keeping generate file   */
    public File getGenSourceDir() {
        return new File(getOutputRoot(), SOURCE_DIR);
    }
    
    /** Get the class name as a path */
    public String getClassNameAsPath() {
        return getTestClass().replace('.', File.separatorChar);
    }

    /** Get the name of the generated source file for this test */
    public File getGenSrcFile() {
        return new File(getGenSourceDir(), getClassNameAsPath() + ".java");
    }

    /** Get the name of the generated class file for this test */
    public File getGenClassFile() {
        return new File(getClassRoot(), getClassNameAsPath() + ".class");
    }

    /** Remove a file, making sure it was deleted */
    public void removeFile(File file) {
        file.delete();
        if (file.exists()) {
            throw new TestException("file still exists after deleting: "
                                    + file);
        }
    }

    /** Check that a file exists file, generate error if it doesn't */
    public void verifyExists(File file) {
        if (!file.exists()) {
            throw new TestException("file was not created: " + file);
        }
    }

    /**
     * Load a document with the XMLC factory, creating one if one is not
     * defined in the class.
     */
    private XMLObject loadDocWithFactory(Object spec)
        throws IOException, ClassNotFoundException {
        XMLCFactory factory = fXMLCFactory;
        if (factory == null) {
            // Create a new factory
            factory = createFactory(getTestClassLoader(), null);
        }
        if (spec instanceof String) {
            return factory.create((String)spec);
        } else {
            return factory.create((Class)spec);
        }
    }

    /**
     * Get a document test class.  This returns the class, not the
     * document object.
     */
    public Class loadDocClass(String className) {
        try {
            return getTestClassLoader().loadClass(className);
        } catch (ClassNotFoundException except) {
            throw new TestException(except);
        }
    }
    
    /**
     * Get the document class.
     */
    public Class loadDocClass() {
        return loadDocClass(getTestClassImpl());
    }
    
    /**
     * Get the document interface.
     */
    public Class loadDocClassInterface() {
        return loadDocClass(getTestClassInterface());
    }
    
    /**
     * Load a document with a classloader.  This simulates a new 
     * in compiled code.
     */
    private XMLObject loadDocWithClassLoader(Object spec)
        throws IllegalAccessException, InstantiationException,
               IOException, ClassNotFoundException {
        Class objClass = getTestClassLoader().loadClass((String)spec);
        return (XMLObject)objClass.newInstance();
    }

    /** 
     * Get an instance of a XMLC document class.
     *
     * @param spec Can be either the name of the class name, interface name
     *  or inteface object.
     */
    public XMLObject loadTestDocument(Object spec) {
        try {
	    return loadDocWithFactory(spec);
        } catch (Throwable except) {
            // Catch everything, this is a common failure point
            throw new TestException("failed to load XMLC document class: \""
                                    + spec + "\"", except);
        }
    }

    /** 
     * Get an instance of the XMLC document class defined by the 
     * current test id.
     */
    public XMLObject loadTestDocument() {
        return loadTestDocument(getTestClass());
    }

    /** 
     * Compile and load a test document
     * @param unqualClassName maybe null
     * @param impls Interface to implement, maybe null.
     */
    public XMLObject compileLoadDoc(File srcFile,
                                    String unqualClassName,
                                    Class impls) {
        // FIXME: could this be replaced with TEstDOMSource factory>>
        ExecXmlc execXmlc = new ExecXmlc(this);
        String className = (unqualClassName != null)
            ? (getTestPackage() + "." + unqualClassName)
            : getTestClass();
        execXmlc.addOpt(ExecXmlc.OPT_CLASS, className);
        if (impls != null) {
            execXmlc.addOpt(ExecXmlc.OPT_IMPLEMENTS,
                            impls.getClass().getName());
        }
        execXmlc.setSrcFile(srcFile);
        execXmlc.compile(getResultFile(COMPILE_OUTPUT_EXT));
        return loadTestDocument(className);
    }

    /**
     * Dump the DOM to a file and verify against an expected.
     */
    public void dumpVerifyDom(Document doc,
                              File resultFile,
                              File expectFile) {
        OutputDocument.dump(doc, resultFile);
        getDiffer().diff(expectFile, resultFile);
    }

    /**
     * Dump the DOM to a file and verify against an expected.
     */
    public void dumpVerifyDom(Document doc,
                              String resultFileExt) {
        dumpVerifyDom(doc, getResultFile(resultFileExt),
                      getExpectedFile(resultFileExt));
    }

    /**
     * Dump the DOM without verification
     */
    public void dumpDom(Document doc,
                        String resultFileExt) {
        OutputDocument.dump(doc, getResultFile(resultFileExt));
    }

    /**
     * Parse a document with XMLC and dump the DOM to a file and verify
     * against an expected.
     */
    public void parseVerifyDom(File srcDoc) {
        // Use XMLC to parse and dump the DOM.
        ExecXmlc dumper = new ExecXmlc(this);
        dumper.addOpt(ExecXmlc.OPT_DUMP);
        dumper.addOpt(ExecXmlc.OPT_METHODS);
        dumper.addOpt(ExecXmlc.OPT_NO_COMPILE);
        dumper.addOpt(ExecXmlc.OPT_CLASS, getTestClass());
        dumper.setSrcFile(srcDoc);
        dumper.compile(getResultFile(PARSE_DOM_DUMP_EXT));
        getDiffer().diff(getExpectedFile(PARSE_DOM_DUMP_EXT),
                         getResultFile(PARSE_DOM_DUMP_EXT));
    }

    /**
     * Get the value of a string field.
     * @param obj is either the object or a class object to get a static
     *  fiels.
     */
    public String getStringField(Object obj,
                                 String fieldName) {
        try {
            if (obj instanceof Class) {
                Field field = ((Class)obj).getField(fieldName);
                return (String)field.get(null);
            } else {
                Field field = obj.getClass().getField(fieldName);
                return (String)field.get(obj);
            }
        } catch (Throwable except) {
            Class cl = (obj instanceof Class) ? (Class)obj
                : obj.getClass();
            throw new TestException("error getting field \"" + fieldName
                                    + "\" of class \"" + cl);
        }
    }

    /** Verify the return type of a method */
    private void verifyReturnType(Method method,
                                  Class returnInstanceOf) {
        if (!returnInstanceOf.isAssignableFrom(method.getReturnType())) {
            throw new TestException("return type of \"" + method.getName()
                                    + "\", \"" + method.getReturnType()
                                    + "\", is not an instance of \""
                                    + returnInstanceOf + "\"");
        }
    }

    /** 
     * Call a getElement method via reflection.
     * @param nameSuffix Is the method name less the `getElement'.
     * @param returnInstanceOf If not null, verify that the return type
     *  is an instance of this class.
     */
    public Element callGetElement(XMLObject xmlObj,
                                  String nameSuffix,
                                  Class returnInstanceOf) {
        String methodName = "getElement" + nameSuffix;
        try {
            Method method = xmlObj.getClass().getMethod(nameSuffix, NO_PARAMS);
            if (returnInstanceOf != null) {
                verifyReturnType(method, returnInstanceOf);
            }
            return (Element)method.invoke(xmlObj, NO_ARGS);
        } catch (Throwable except) {
            throw new TestException("error calling \"" + methodName
                                    + "\" of class \""
                                    + xmlObj.getClass().getName());
        }
    }

    /** 
     * Call a setTest method via reflection.
     * @param nameSuffix Is the method name less the `setText'.
     */
    public void callSetText(XMLObject xmlObj,
                            String nameSuffix,
                            String value) {
        String methodName = "setText" + nameSuffix;
        try {
            Method method = xmlObj.getClass().getMethod(nameSuffix, STR_PARAM);
            verifyReturnType(method, Void.class);
            method.invoke(xmlObj, new String[]{value});
        } catch (Throwable except) {
            throw new TestException("error calling \"" + methodName
                                    + "\" of class \""
                                    + xmlObj.getClass().getName());
        }
    }

    /** Load a class by name */
    private Class loadClass(String name) {
        try {
            return getClass().getClassLoader().loadClass(name);
        } catch (ClassNotFoundException except) {
            throw new TestException(except);
        }
    }

    /** check is a class is an instance of a specified class */
    public void verifyInstanceOf(Class testClass,
                                 Class expectClass) {
        if (!expectClass.isAssignableFrom(testClass)) {
            throw new TestException(testClass + " is not an instance of "
                                    + expectClass);
        }
    }

    /** check is a class is an instance of a specified class name */
    public void verifyInstanceOf(Class testClass,
                                 String expectClass) {
        verifyInstanceOf(testClass, loadClass(expectClass));
    }

    /** check is a class is an instance of all classes in a list */
    public void verifyInstanceOf(Class testClass,
                                 String[] expectClasses) {
        for (int i = 0; i < expectClasses.length; i++) {
            verifyInstanceOf(testClass, expectClasses[i]);
        }
    }

    /** check is a class is not an instance of a specified class */
    public void verifyNotInstanceOf(Class testClass,
                                    Class expectClass) {
        if (expectClass.isAssignableFrom(testClass)) {
            throw new TestException(testClass
                                    + " should not be an instance of "
                                    + expectClass);
        }
    }

    /** check is a class is not an instance of a specified class name */
    public void verifyNotInstanceOf(Class testClass,
                                    String expectClass) {
        verifyNotInstanceOf(testClass, loadClass(expectClass));
    }

    /** check is a class is not an instance of all classes in a list */
    public void verifyNotInstanceOf(Class testClass,
                                    String[] expectClasses) {
        for (int i = 0; i < expectClasses.length; i++) {
            verifyNotInstanceOf(testClass, expectClasses[i]);
        }
    }

    /* Verify that a class object is an interface */
    public void verifyIsInterface(Class testClass) {
        if (!testClass.isInterface()) {
            throw new TestException("class is not an interface:"
                                    + testClass);
        }
    }

    /* Verify that a class object is an class (not an interface) */
    public void verifyIsClass(Class testClass) {
        if (testClass.isInterface() || testClass.isPrimitive()) {
            throw new TestException("class is not an class:"
                                    + testClass);
        }
    }
}
