/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ShelfTests.java,v 1.1.1.1 2003/03/10 16:36:28 taweili Exp $
 */

package org.enhydra.xml.xmlc.html;

import org.enhydra.xml.xmlc.driver.XmlcTestParams;
import org.enhydra.xml.xmlc.driver.ExecXmlc;
import org.enhydra.xml.xmlc.driver.TestDOMSource;
import org.enhydra.xml.dom.DOMInfo;
import org.enhydra.xml.xmlc.XMLCUtil;
import java.net.URLEncoder;
import org.w3c.dom.Node;
import org.w3c.dom.Text;
import org.w3c.dom.html.HTMLParagraphElement;
import org.w3c.dom.html.HTMLAnchorElement;
import org.w3c.dom.html.HTMLImageElement;
import org.w3c.dom.html.HTMLParagraphElement;
import org.w3c.dom.html.HTMLAnchorElement;
import java.util.Vector;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Enumeration;
import java.lang.reflect.Method;
import junit.framework.Test;

/**
 * Test case, based on the shopping demo, that does some manipulation of the
 * DOM, including cloning and inserting.
 */
public class ShelfTests extends HtmlTestCaseBase {
    /**
     * Compile-time debug flag.
     */
    private static boolean DEBUG = false;

    /**
     * Wrapper around XMLC factory object, includes compilation on first
     * use.
     */
    private TestDOMSource fShelfSource;
    
    /**
     * Root of similated Inventory.
     */
    private static Inventory fInventoryRoot;

    /**
     * Current request number for DOM dumps.
     */
    private static int fRequestCnt;

    /**
     * URL to reference ourselve.
     */
    private static final String SHELF_URL = "Shelf.po";

    /**
     * URL to reference the item presentation object.
     */
    private static final String ITEM_URL = "Item.po";

    /**
     * Image to use for open categories.
     */
    private static final String OPEN_CAT_IMAGE = "../media/true.gif";
    private static final String CLOSED_CAT_IMAGE = "../media/false.gif";

    /**
     * Class for generating a simulating a request.
     */
    private class RequestTest {
        /**
         * The catagory id of the button that was pressed or NULL if
         * one was not pressed.
         */
        private Long fButtonCatId;

        /**
         * Table of category ids that are open.  Each entry contains a Long
         * object.
         */
        private Hashtable fOpenCatIds = new Hashtable();

        /**
         * CGI argument containing all open category ids.  If null,
         * then none are open.
         */
        private String fOpenCatCgiArg;

        /**
         * Shelf HTML object, addressed as an interface.
         */
        private ShelfHTML fShelfHtml;

        /**
         * Pointers to category prototype entries.
         */
        private HTMLParagraphElement fSampleCatBlock;
        private HTMLAnchorElement fSampleCatAnchor;
        private HTMLImageElement fSampleCatImg;
        private Text fSampleCatText;

        /**
         * Pointers to item prototype entries.
         */
        private HTMLParagraphElement fSampleItemBlock;
        private HTMLAnchorElement fSampleItemAnchor;
        private Text fSampleItemText;

        /** Constructor */
        public RequestTest() { 
        }

        /**
         * Remove prototype table and save pointers to the prototype
         * entries.  For easy of use, prototypes are edited and then
         * cloned.
         */
        private void prepareHtml() {
            fShelfHtml = (ShelfHTML)fShelfSource.create();

            // Save pointers to prototypes, remove ids as they shouldn't be duplicated.
            fSampleCatBlock = fShelfHtml.getElementSampleCatBlock();
            fSampleCatAnchor = fShelfHtml.getElementSampleCatAnchor();
            fSampleCatImg = fShelfHtml.getElementSampleCatImg();
            fSampleCatText = XMLCUtil.getFirstText(fShelfHtml.getElementSampleCatText());

            fSampleItemBlock = fShelfHtml.getElementSampleItemBlock();
            fSampleItemAnchor = fShelfHtml.getElementSampleItemAnchor();
            fSampleItemText = XMLCUtil.getFirstText(fShelfHtml.getElementSampleItemText());

            // Remove children of prototype list.
            Node listNode = fShelfHtml.getElementList();
            Node child = listNode.getFirstChild();
            while (child != null) {
                Node nextChild = child.getNextSibling();
                listNode.removeChild(child);
                child = nextChild;
            }

        }

        /**
         * Create a new HTML entry for a category.
         */
        private void makeCategoryHtml(Inventory category, Node parent,
                                      boolean isOpen, int level) {
            if (DEBUG) {
                msgPrintln("makeCategotyHtml enter: " + level);
            }
            long id = category.getObjectId();

            String url = SHELF_URL + "?button=" + id;
            if (fOpenCatCgiArg != null) {
                url += "&" + fOpenCatCgiArg;
            }
            fSampleCatAnchor.setHref(url);
            if (isOpen) {
                fSampleCatImg.setSrc(OPEN_CAT_IMAGE);
            } else {
                fSampleCatImg.setSrc(CLOSED_CAT_IMAGE);
            }
            fSampleCatText.setData(category.getDesc());
            fSampleCatBlock.setClassName("level" + level);
            parent.appendChild(fSampleCatBlock.cloneNode(true));
            if (DEBUG) {
                msgPrintln("makeCategotyHtml exit: " + level);
            }
        }

        /**
         * Create a new HTML entry for an item.
         */
        private void makeItemHtml(Inventory item, Node parent, int level) {
            if (DEBUG) {
                msgPrintln("makeItemHtml enter: " + level);
            }
            String url = ITEM_URL + "?ref=../contents/" + item.getRef()
                + "&add=" + item.getObjectId();
            fSampleItemAnchor.setHref(url);
            fSampleItemAnchor.setTarget("item");
            fSampleItemText.setData(item.getDesc());
            fSampleItemBlock.setClassName("level" + level);
            parent.appendChild(fSampleItemBlock.cloneNode(true));
            if (DEBUG) {
                msgPrintln("makeItemHtml exit: " + level);
            }
        }

        /**
         * Recursively process a category or item entry, adding it to the DOM.
         * Sets node class using level for CSS.
         */
        private void processInventoryEntry(Inventory entry, Node parent, int level) {
            if (entry.isCategory()) {
                boolean isOpen = fOpenCatIds.contains(new Long(entry.getObjectId()));
                makeCategoryHtml(entry, parent, isOpen, level);
                if (isOpen) {
                    Vector entries = entry.getChildren();
                    for (int idx = 0; idx < entries.size(); idx++) {
                        processInventoryEntry((Inventory)entries.elementAt(idx), parent, level+1);
                    }
                }
            } else {
                makeItemHtml(entry, parent, level);
            }
        }

        /**
         * Generate the DOM entries for the current state of the inventory.
         */
        private void processInventory() {
            processInventoryEntry(fInventoryRoot,
                                  fShelfHtml.getElementList(), 1);
        }

        /**
         * Parse CGI arguments into fields.
         */
        private void parseArgs(String button,
                               String open) {

            // Parse button id.
            if (button != null) {
                fButtonCatId = new Long(button);
            }

            if (open != null) {
                StringTokenizer tokens = new StringTokenizer(open, ",");
                while (tokens.hasMoreTokens()) {
                    String token = tokens.nextToken();
                    try {
                        Long id = new Long(token);
                        fOpenCatIds.put(id, id);
                    } catch (NumberFormatException except) {
                        // Ignore, probably a bogus URL.
                    }
                }
            }
        }

        /**
         * Modify the open category list based on the currenmt button.
         * Also generate CGI argument for the open categories.
         */
        private void adjustOpenCategories() {
            if (fButtonCatId != null) {
                // Adjust the open entry for the current button.
                if (fOpenCatIds.containsKey(fButtonCatId)) {
                    fOpenCatIds.remove(fButtonCatId);
                } else {
                    fOpenCatIds.put(fButtonCatId, fButtonCatId);
                }
            }

            // Generate CGI arguments.
            StringBuffer buf = new StringBuffer();
            Enumeration openIds = fOpenCatIds.keys();

            while (openIds.hasMoreElements()) {
                if (buf.length() > 0) {
                    buf.append(",");
                }
                buf.append(((Long)openIds.nextElement()).toString());
            }
            if (buf.length() > 0) {
                fOpenCatCgiArg = "open=" + buf.toString();
            }
        }

        /**
         * Run a simulated request.
         */
        public void runRequest(String button,
                               String open) {
            parseArgs(button, open);
            prepareHtml();
            adjustOpenCategories();
            processInventory();
        }

    }

    /** Factory method to create suite of these tests */
    public static Test suite() {
        return createSuite(ShelfTests.class, null);
    }
    
    /** Constructor */
    public ShelfTests(Method method) {
        super(method);
        fShelfSource = new TestDOMSource(this, getParams(),
                                         getInputFile("Shelf.html"),
                                         "ShelfHTMLImpl",
                                         ShelfHTML.class);
    }

    /**
     * Create an object and simulate a request.
     */
    private void request(String button,
                         String open) {
        RequestTest test = new RequestTest();
        test.runRequest(button, open);
        fRequestCnt++;
    }

    /**
     * Playback recorded http requests.
     */
    private void playback() {
        request(null, null);
        request("0", null);
        request("2", "0");
        request("2", "2,0");
        request("7", "0");
        request("8", "7,0");
        request("9", "8,7,0");
        request("7", "9,8,7,0");
        request("7", "9,8,0");
        request("8", "9,8,7,0");
        request("7", "9,7,0");
        request("3", "9,0");
        request("4", "9,3,0");
        request("1", "9,4,3,0");
        request("2", "9,4,3,1,0");
        request("1", "9,4,3,2,1,0");
        request("0", "9,4,3,2,0");
        request("0", "9,4,3,2");
        request("2", "9,4,3,2,0");
        request("4", "9,4,3,0");
        request("3", "9,3,0");
        request("1", "9,0");
        request("1", "9,1,0");
        request("0", "9,0");
        request("0", "9");
        request("7", "9,0");
        request("3", "9,7,0");
        request("2", "9,7,3,0");
        request("1", "9,7,3,2,0");
        request("0", "9,7,3,2,1,0");
        request("0", "9,7,3,2,1");
        request("1", "9,7,3,2,1,0");
        request("2", "9,7,3,2,0");
        request("5", "9,7,3,0");
        request("2", "9,7,5,3,0");
        request("7", "9,7,5,3,2,0");
        request("2", "9,5,3,2,0");
    }
    
    /**
     * Build test inventory (build from tracing demo program).
     */
    private Inventory buildInventory() {
        Inventory entry0, entry1, entry2, entry3;
        entry0 = new Inventory(true, 0, "Store Directory", "Main Menu");
        entry1 = new Inventory(true, 1, "Accessories", "Accessories");
        entry0.addChild(entry1);
        entry2 = new Inventory(false, 5, "Assorted Balls", "pages/accessories_balls.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 6, "Gloves", "pages/accessories_gloves.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 7, "Practice Balls", "pages/accessories_practiceballs.html");
        entry1.addChild(entry2);
        entry1 = new Inventory(true, 2, "Clubs", "Clubs");
        entry0.addChild(entry1);
        entry2 = new Inventory(false, 8, "Graphite Irons", "pages/clubs_irons_graphite.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 9, "Black Hawk Putter", "pages/clubs_putter_1.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 10, "Steel Irons", "pages/clubs_irons_steel.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 11, "Basic Putter", "pages/clubs_putter_2.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 12, "Grey Hawk Putter", "pages/clubs_putter_3.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 13, "Side Waggle Putter", "pages/clubs_putter_4.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 14, "2 Tone Putter", "pages/clubs_putter_5.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 15, "Graphite Wood", "pages/clubs_woods_graphite.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 16, "Steel Wood Driver", "pages/clubs_woods_steel.html");
        entry1.addChild(entry2);
        entry2 = new Inventory(false, 17, "Wood Woods", "pages/clubs_woods_woods.html");
        entry1.addChild(entry2);
        entry1 = new Inventory(true, 3, "Men's Apparel", "MensApparel");
        entry0.addChild(entry1);
        entry2 = new Inventory(true, 4, "Shirts", "Shirts");
        entry1.addChild(entry2);
        entry3 = new Inventory(false, 22, "Classic Golf Shirt", "pages/mens_shirt_1.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 23, "Pocketed Golf Shirt", "pages/mens_shirt_2.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 24, "No-Pocket Polo", "pages/mens_shirt_3.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 25, "Striped Polo", "pages/mens_shirt_4.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 26, "Long Sleeve ", "pages/mens_shirt_5.html");
        entry2.addChild(entry3);
        entry2 = new Inventory(true, 5, "Shoes", "Shoes");
        entry1.addChild(entry2);
        entry3 = new Inventory(false, 27, "Golf-Classic Series", "pages/mens_shoes_1.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 28, "Talon Golf Shoe", "pages/mens_shoes_2.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 29, "Sport Golf Shoe", "pages/mens_shoes_3.html");
        entry2.addChild(entry3);
        entry2 = new Inventory(true, 6, "Sweaters", "Sweaters");
        entry1.addChild(entry2);
        entry3 = new Inventory(false, 30, "Pretty-N-Pink Sweater", "pages/mens_sweater_1.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 31, "Stylish Sweater", "pages/mens_sweater_2.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 32, "Supplex Wind Shirt", "pages/mens_sweater_3.html");
        entry2.addChild(entry3);
        entry1 = new Inventory(true, 7, "Women's Apparel", "WomensApparel");
        entry0.addChild(entry1);
        entry2 = new Inventory(true, 8, "Shirts", "Shirts");
        entry1.addChild(entry2);
        entry3 = new Inventory(false, 33, "Short Sleeve Polo", "pages/womens_shirt_1.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 34, "Striped Polo Jersey", "pages/womens_shirt_2.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 35, "Fancy Collar Jersey", "pages/womens_shirt_3.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 36, "Polo Jersey", "pages/womens_shirt_4.html");
        entry2.addChild(entry3);
        entry2 = new Inventory(true, 9, "Shorts", "Shorts");
        entry1.addChild(entry2);
        entry3 = new Inventory(false, 39, "Microfiber Twill", "pages/womens_shorts_1.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 40, "Twill II", "pages/womens_shorts_2.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 41, "Plaid Twill", "pages/womens_shorts_3.html");
        entry2.addChild(entry3);
        entry3 = new Inventory(false, 42, "Khaki Shorts", "pages/womens_shorts_4.html");
        entry2.addChild(entry3);
        return entry0;
    }

    /**
     * Run the test.
     */
    public void test1() {
        fInventoryRoot = buildInventory();
        playback();
    }
}

/**
 * Entry in the inventory tree.
 */
class Inventory {
    private boolean fIsCategory;
    private long fObjectId;
    private String fDesc;
    private String fRef;
    private Vector children;

    /*
     * Constructor
     */ 
    public Inventory(boolean isCategory,
                     long objectId,
                     String desc,
                     String ref) {
        fIsCategory = isCategory;
        fObjectId = objectId;
        fDesc = desc;
        fRef = ref;
    }

    public void addChild(Inventory child) {
        if (children == null) {
            children = new Vector();
        }
        children.add(child);
    }

    public Vector getChildren() {
        return children;
    }

    public boolean isCategory() {
        return fIsCategory;
    }

    public long getObjectId() {
        return fObjectId;
    }

    public String getDesc() {
        return fDesc;
    }

    public String getRef() {
        return fRef;
    }
}
