/*
 * Enhydra Java Application Server Project
 * 
 * The contents of this file are subject to the Enhydra Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License on
 * the Enhydra web site ( http://www.enhydra.org/ ).
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See 
 * the License for the specific terms governing rights and limitations
 * under the License.
 * 
 * The Initial Developer of the Enhydra Application Server is Lutris
 * Technologies, Inc. The Enhydra Application Server and portions created
 * by Lutris Technologies, Inc. are Copyright Lutris Technologies, Inc.
 * All Rights Reserved.
 * 
 * Contributor(s):
 * 
 * $Id: ExtendsTests.java,v 1.1.1.1 2003/03/10 16:37:12 taweili Exp $
 */

package org.enhydra.xml.xmlc.xml;
import java.io.File;
import org.enhydra.xml.driver.TestDiff;
import org.enhydra.xml.xmlc.driver.XmlcTestParams;
import org.enhydra.xml.xmlc.driver.ExecXmlc;
import org.enhydra.xml.xmlc.XMLObject;
import java.lang.reflect.Method;
import junit.framework.Test;

/**
 * Tests of generating classes where super class or interfaces are specified.
 * Also tests of -domfactory
 */
public class ExtendsTests extends XmlTestCaseBase {
    /** Interfaces a generated interface must be an instanceof */
    private final String[] INTERFACE_INSTANCEOF = {
	"org.enhydra.xml.xmlc.XMLObject",
	"org.w3c.dom.Document"
    };

    /** Interfaces a generated interface must not be an instanceof */
    private final String[] INTERFACE_NOT_INSTANCEOF = {
	"org.enhydra.xml.xmlc.XMLObjectImpl",
    };

    /** Interfaces/classes a generated class must be an instanceof */
    private final String[] IMPLEMENTATION_INSTANCEOF = {
	"org.enhydra.xml.xmlc.XMLObject",
	"org.enhydra.xml.xmlc.XMLObjectImpl",
	"org.w3c.dom.Document"
    };

    /** Factory method to create suite of these tests */
    public static Test suite() {
        return createSuite(ExtendsTests.class, null);
    }
    
    /** Constructor */
    public ExtendsTests(Method method) {
        super(method);
    }

    /**
     * Do standard verification for an generate class.
     */
    private void verifyClass(XMLObject doc,
                             Class xtraImpl) {
        verifyIsClass(doc.getClass());
        verifyInstanceOf(doc.getClass(),
                              IMPLEMENTATION_INSTANCEOF);
        if (xtraImpl != null) {
            verifyInstanceOf(doc.getClass(), xtraImpl);
        }
    }

    /**
     * Do standard verification for an generate interface.
     */
    private void verifyInterface(Class docClass,
                                 Class xtraImpl) {
        verifyIsInterface(docClass);
        verifyInstanceOf(docClass,
                         INTERFACE_INSTANCEOF);
        verifyNotInstanceOf(docClass,
                            INTERFACE_NOT_INSTANCEOF);
        if (xtraImpl != null) {
            verifyInstanceOf(docClass, xtraImpl);
        }
    }

    /**
     * Craete the test class used in most of these tests
     */
    private XmlBasicTest createTest() {
        return new XmlBasicTest(this, getInputFile("wml/Login.wml"),
                                getInputFile("wml/wml_1.1.test.xml"));
    }

    /*
     * Test 1: -implements
     */
    public void test1() {
        XmlBasicTest test = createTest();
        test.addFirst(ExecXmlc.OPT_IMPLEMENTS, WMLLogin.class.getName());
        
        Class expectDocClass;
        if (fParams.getDom().equals(ExecXmlc.XERCES_DOM)) {
            test.setDomOpt(ExecXmlc.OPT_DOM_FACTORY,
                           TestXercesDOMFactory.class.getName());
            expectDocClass = TestXercesDOMDocument.class;
        } else {
            test.setDomOpt(ExecXmlc.OPT_DOM_FACTORY,
                           TestLazyDOMFactory.class.getName());
            expectDocClass = TestLazyDOMDocument.class;
        }
        test.basicTest();

        XMLObject doc = loadTestDocument();
        verifyClass(doc, WMLLogin.class); 
        verifyInstanceOf(doc.getDocument().getClass(),
                         expectDocClass);
    }

    /*
     * Test 2: -extends
     */
    public void test2() {
        XmlBasicTest test = createTest();
        test.addFirst(ExecXmlc.OPT_EXTENDS, XMLObjectBase.class.getName());
        
        Class expectDocClass;
        if (fParams.getDom().equals(ExecXmlc.XERCES_DOM)) {
            test.setDomOpt(ExecXmlc.OPT_DOM_FACTORY,
                           TestXercesDOMFactory.class.getName());
            expectDocClass = TestXercesDOMDocument.class;
        } else {
            test.setDomOpt(ExecXmlc.OPT_DOM_FACTORY,
                           TestLazyDOMFactory.class.getName());
            expectDocClass = TestLazyDOMDocument.class;
        }
        test.basicTest();

        XMLObject doc = loadTestDocument();
        verifyClass(doc, XMLObjectBase.class); 
        verifyInstanceOf(doc.getDocument().getClass(),
                         expectDocClass);
    }

    /*
     * Test 3:  generating an interfaces and an implementation, and
     * implementing an interface.
     */
    public void test3() {
        XmlBasicTest test = createTest();
        setUsingImplInterface();
        test.addFirst(ExecXmlc.OPT_GENERATE, ExecXmlc.GENERATE_BOTH);
        test.addFirst(ExecXmlc.OPT_IMPLEMENTS,
                      WMLLogin.class.getName());
        test.basicTest();
        
        verifyInterface(loadDocClassInterface(), WMLLogin.class);
        verifyClass(loadTestDocument(), WMLLogin.class);
    }

    /*
     * Test 4: generating an interfaces and an implementation, extending
     * a base class
     */
    public void test4() {
        XmlBasicTest test = createTest();
        setUsingImplInterface();
        test.addFirst(ExecXmlc.OPT_GENERATE, ExecXmlc.GENERATE_BOTH);
        test.addFirst(ExecXmlc.OPT_EXTENDS,
                      XMLObjectBase.class.getName());
        test.basicTest();
        
        verifyInterface(loadDocClassInterface(), null);
        verifyClass(loadTestDocument(), XMLObjectBase.class); 
    }

    /*
     * Test 5: - not sure what was special about this test
     */
    public void test5() {
        XmlBasicTest test = createTest();
        setUsingImplInterface();
        test.addFirst(ExecXmlc.OPT_EXTENDS,
                      XMLObjectBase.class.getName());
        test.addFirst(ExecXmlc.OPT_GENERATE, ExecXmlc.GENERATE_BOTH);
        test.basicTest();
        verifyInterface(loadDocClassInterface(), null);
        verifyClass(loadTestDocument(), XMLObjectBase.class); 
    }

    /*
     * Test 6: generate an interface and an implementation seperately
     */
    public void test6() {
        XmlBasicTest test
            = new XmlBasicTest(this, getInputFile("bioml/insulin3.bioml"),
                               getInputFile("bioml/bioml.dtd"));
        setUsingImplInterface();
        
        // Compile the interface
        test.addFirst(ExecXmlc.OPT_GENERATE, ExecXmlc.GENERATE_INTERFACE);
        test.compileOriginalStep();

        // now, compile the implementation
        ExecXmlc execXmlc = new ExecXmlc(this);
        execXmlc.addOpt(ExecXmlc.OPT_GENERATE,
                        ExecXmlc.GENERATE_IMPLEMENTATION);
        execXmlc.addOpt(ExecXmlc.OPT_CLASS, getTestClass());
        execXmlc.addOpt(ExecXmlc.OPT_DEST_DIR, getClassRoot());
        execXmlc.addClassPath(getClassRoot());
        execXmlc.setSrcFile(getInputFile("bioml/insulin3.bioml"));

        File outFile2 = getResultFile("compile2.out");
        File expectFile2 = getExpectedFile("compile2.out");
        execXmlc.compile(outFile2);
        getDiffer().diff(expectFile2, outFile2);
        
        // Verify interface and implementation.  The test object used to
        // compile the interface is used to load both so that we get the
        // same classloader.
        Class faceClass = loadDocClassInterface();
        verifyInterface(faceClass, null);
        verifyClass(loadTestDocument(faceClass), faceClass); 

        // finish test now that interface is compiled
        test.reparseStep();
    }
}
